module hdesim3
  use constants; use mympitools; use mesh; use division; use mylinsol
  implicit none
  private
  public :: hdesim3main

contains
  subroutine hdesim3main(ul,vl,wl,pl,viscl,dtrml,ivlt,ivlb, &
       ivis,relu,relv,relw,relp,isolvers)
    !***********************************************************************
    !     solve three-dimensional hydrodynamic equations
    !     by using SIMPLER method
    !     harmonic equations are solved by (preconditioned) CG method
    !     by calling subroutine cgbase123p
    !
    !     20030625: convergency check is improved
    !***********************************************************************

    integer :: i,j,k,ij,ich,nsmp,maxitr
    integer,parameter :: maxsmp=9999
    integer,intent(in) :: ivlt,ivlb,ivis
    real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: &
         ul,vl,wl,pl,viscl,dtrml

    ! ----------------------------------------
    !     variables for SIMPLER iterations
    ! ----------------------------------------

    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         :: vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
         ,dtrmwk
    real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: divv,buoy
    real(dp),dimension(7,ista:iend,jsta:jend,ksta:kend) :: &
         apx1,apy1,apz1,app1,apx2,apy2,apz2,app2
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
         dudp1,dvdp1,dwdp1
    real(dp) :: zansa,errors(maxsmp)
    real(dp),parameter :: epssim=1.e-6_dp
    real(dp) :: vnorm2,divnorm2,vdhnorm2,vnorm2p,divnorm2p,vdhnorm2p
    real(dp) :: divzansa,vdhzansa
    real(dp) :: rtmp1(3),rtmp2(3*nprocs)
    real(dp),intent(in) :: relu,relv,relw,relp
    real(dp) :: arelu,arelv,arelw,arelp,arelu1,arelv1,arelw1,arelp1
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
         pwk,pdash,uhat,uwk,uold,vhat,vwk,vold,what,wwk,wold
    real(dp) :: udash2,vdash2,wdash2

    ! -------------------------------------
    !     variables for parallelization
    ! -------------------------------------

    integer :: istau,iendu,ista2u,iend1u
    integer :: jstav,jendv,jsta2v,jend1v
    integer :: kstaw,kendw,ksta2w,kend1w
    ! -----------------------------
    !     variables for solver
    ! -----------------------------

    integer :: nitr,istat,isolvers(6)
    real(dp) :: eitr
    real(dp),parameter :: epsres=1.e-8_dp !1.e-15_dp
    character(4) :: prec_name

    integer,parameter :: mxx=iax-3, myx=jax-2, mzx=kax-2, nmx=mxx*myx*mzx
    real(dp),allocatable :: amtx(:,:,:,:),bvcx(:,:,:),xvcx(:,:,:) &
         ,amtxpc(:,:,:,:),diagix(:,:,:)
    integer iprecx

    integer,parameter :: mxy=iax-2, myy=jax-3, mzy=kax-2, nmy=mxy*myy*mzy
    real(dp),allocatable :: amty(:,:,:,:),bvcy(:,:,:),xvcy(:,:,:) &
         ,amtypc(:,:,:,:),diagiy(:,:,:)
    integer iprecy

    integer,parameter :: mxz=iax-2, myz=jax-2, mzz=kax-3, nmz=mxz*myz*mzz
    real(dp),allocatable :: amtz(:,:,:,:),bvcz(:,:,:),xvcz(:,:,:) &
         ,amtzpc(:,:,:,:),diagiz(:,:,:)
    integer iprecz

    integer,parameter :: mxp=iax-2, myp=jax-2, mzp=kax-2, nmp=mxp*myp*mzp
    real(dp),allocatable :: amtp(:,:,:,:),bvcp(:,:,:),xvcp(:,:,:) &
         ,amtppc(:,:,:,:),diagip(:,:,:)
    integer iprecp

!!$    maxitr = maxsmp
!!$  if (ivis == 0) then
    if (ivis**2+ivlt**2+ivlb**2 == 0) then
       maxitr = 1
    else
       maxitr = maxsmp
    endif

! -----------------------------------------
!     calculate buoyancy at grid points
! -----------------------------------------

    dtrmwk = 0.e0_dp
    dtrmwk(ista:iend, jsta:jend, ksta:kend) &
         = dtrml(ista:iend, jsta:jend, ksta:kend)
    call mypartialshift3l1(dtrmwk,ista,iend,jsta,jend,ksta,kend)
    do k = ksta2,kend
       do j = jsta2,jend1
          do i = ista2,iend1
             buoy(i,j,k) = 0.5e0_dp*dx(i)*dy(j)* &
                  (dtrmwk(i,j,k-1)*dz(k-1)+dtrmwk(i,j,k)*dz(k))
          enddo
       enddo
    enddo

    ! ------------------------------------------
    !     (for variable viscosity)
    !     calculate viscosity at grid points
    ! ------------------------------------------

    if (ivis /= 0) then
       call viscnd(viscl,vcc,ista,iend,jsta,jend,ksta,kend)
       call myroundshift123l(vcc,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)
       call calvisg(vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
            ,dx,dy,dz,dxs,dys,dzs,dxi,dyi,dzi,dxsi,dysi,dzsi &
            ,ista,iend,jsta,jend,ksta,kend)
    endif

    ! ========================================
    !     calculate coefficients for u/v/w
    ! ========================================

    ! level 1: discretization
    call make_simpler_coefs_level1 &
         (ivis,apx1,apy1,apz1,app1,dudp1,dvdp1,dwdp1 &
         ,vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
         ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
         ,ista,iend,jsta,jend,ksta,kend)

    ! level 2: boundary conditions imposed
    call make_simpler_coefs_level2 &
         (apx1,apy1,apz1,app1,apx2,apy2,apz2,app2 &
         ,dudp1,dvdp1,dwdp1 &
         ,ista,iend,jsta,jend,ksta,kend &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1,ivlt,ivlb)

    ! -----------------------------------------
    !     set/decompose coefficient matrice
    ! -----------------------------------------

    istau = ista2-1; iendu = iend1-1
    jstav = jsta2-1; jendv = jend1-1
    kstaw = ksta2-1; kendw = kend1-1
    ista2u = max(2,istau); iend1u = min(mxx+1,iendu)
    jsta2v = max(2,jstav); jend1v = min(myy+1,jendv)
    ksta2w = max(2,kstaw); kend1w = min(mzz+1,kendw)

    allocate &
         (amtx(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
         ,amty(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1,7) &
         ,amtz(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1,7) &
         ,amtp(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
         ,stat=istat)
    call make_simpler_matrice &
         (apx2,apy2,apz2,app2,amtx,amty,amtz,amtp &
         ,ista,iend,jsta,jend,ksta,kend &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w)
    allocate &
         (bvcx(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,xvcx(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,amtxpc(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
         ,diagix(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,stat=istat)
    !  write(*,'("vectors allocation for u",i2)')istat
    allocate &
         (bvcy(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
         ,xvcy(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
         ,amtypc(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1,7) &
         ,diagiy(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
         ,stat=istat)
    !  write(*,'("vectors allocation for v",i2)')istat
    allocate &
         (bvcz(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1) &
         ,xvcz(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1) &
         ,amtzpc(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1,7) &
         ,diagiz(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1) &
         ,stat=istat)
    !  write(*,'("vectors allocation for w",i2)')istat
    allocate &
         (bvcp(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,xvcp(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,amtppc(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
         ,diagip(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,stat=istat)
    !  write(*,'("vectors allocation for p",i2)')istat

    ! -----------------------
    !     set work arrays
    ! -----------------------

    uwk = 0.e0_dp; vwk = 0.e0_dp; wwk = 0.e0_dp
    uhat = 0.e0_dp; vhat = 0.e0_dp; what = 0.e0_dp; pwk = 0.e0_dp
    uwk(ista:iend,jsta:jend,ksta:kend) = ul(ista:iend,jsta:jend,ksta:kend)
    vwk(ista:iend,jsta:jend,ksta:kend) = vl(ista:iend,jsta:jend,ksta:kend)
    wwk(ista:iend,jsta:jend,ksta:kend) = wl(ista:iend,jsta:jend,ksta:kend)
    pwk(ista:iend,jsta:jend,ksta:kend) = pl(ista:iend,jsta:jend,ksta:kend)
    call myroundshift123l(uwk,ista,iend,jsta,jend,ksta,kend &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1)
    call myroundshift123l(vwk,ista,iend,jsta,jend,ksta,kend &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1)
    call myroundshift123l(wwk,ista,iend,jsta,jend,ksta,kend &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1)
    call velcndl(uwk,vwk,wwk,ivlt,ivlb,ista,iend,jsta,jend,ksta,kend)

    ! relaxation is not applied for 1st iteration loop
    arelu = 1.e0_dp; arelu1 = 0.e0_dp
    arelv = 1.e0_dp; arelv1 = 0.e0_dp
    arelw = 1.e0_dp; arelw1 = 0.e0_dp
    arelp = 1.e0_dp; arelp1 = 0.e0_dp

    ! initialize solver
    iprecx = 0; iprecy = 0; iprecz = 0; iprecp = 0

    ! //////////////////////////
    !     simpler loop start
    ! //////////////////////////

    ich = 1
    simplerloop : do nsmp = 1,maxitr

       ! save values at previous iteration
       uold = uwk; vold = vwk; wold = wwk

       ! ----------------------------------------------
       !     step 2: calculate dummy velocity field
       ! ----------------------------------------------

       call do_simpler_step2 &
            (uhat,vhat,what,uwk,vwk,wwk,apx1,apy1,apz1 &
            ,dx,dy,dz,ivis &
            ,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy,buoy &
            ,ista,iend,jsta,jend,ksta,kend)

       call velcndl(uhat,vhat,what,ivlt,ivlb,ista,iend,jsta,jend,ksta,kend)

       ! ------------------------------------
       !     step 3: guess pressure field
       ! ------------------------------------

       call mypartialshift1l2(uhat,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift2l2(vhat,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift3l2(what,ista,iend,jsta,jend,ksta,kend)
       call calc_divvl &
            (uhat,vhat,what,divv,dx,dy,dz,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)
       bvcp = 0.e0_dp; xvcp = 0.e0_dp
       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                bvcp(i,j,k) = divv(i,j,k)
                xvcp(i,j,k) = pwk(i,j,k)
             enddo
          enddo
       enddo
       if (myrank == 0)bvcp(2,2,2) = 0.e0_dp
       eitr = epsres; nitr = nmp
       call cgbase123pl &
            (amtp,bvcp,xvcp,eitr,nitr,istat &
            ,amtppc,diagip,ista2,iend1,jsta2,jend1,ksta2,kend1 &
            ,isolvers(1),iprecp,prec_name)
#ifdef VERBOSE
       if (myrank == 0) then
          write(*,'(a8,2a11)')"pstar","nitr","eitr"
          write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
       endif
#endif

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                pwk(i,j,k) = xvcp(i,j,k)*arelp+pwk(i,j,k)*arelp1
             enddo
          enddo
       enddo

       ! ------------------------------------
       !     step 4: guess velocity field
       ! ------------------------------------

       call mypartialshift1l1(pwk,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift2l1(pwk,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift3l1(pwk,ista,iend,jsta,jend,ksta,kend)
       bvcx = 0.e0_dp; xvcx = 0.e0_dp
       bvcy = 0.e0_dp; xvcy = 0.e0_dp
       bvcz = 0.e0_dp; xvcz = 0.e0_dp
       call prep_simpler_step4 &
            (bvcx,bvcy,bvcz,uwk,vwk,wwk,dx,dy,dz,ivis &
            ,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy,buoy,pwk &
            ,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
            ,istau,iendu,jstav,jendv,kstaw,kendw &
            ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w)

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista3,iend1
                xvcx(i-1,j,k) = uwk(i,j,k)
             enddo
          enddo
       enddo
       do k = ksta2,kend1
          do j = jsta3,jend1
             do i = ista2,iend1
                xvcy(i,j-1,k) = vwk(i,j,k)
             enddo
          enddo
       enddo
       do k = ksta3,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                xvcz(i,j,k-1) = wwk(i,j,k)
             enddo
          enddo
       enddo

       eitr = epsres; nitr = nmx
       call cgbase123pl &
            (amtx,bvcx,xvcx,eitr,nitr,istat &
            ,amtxpc,diagix,ista2u,iend1u,jsta2,jend1,ksta2,kend1 &
            ,isolvers(2),iprecx,prec_name)
#ifdef VERBOSE
       if (myrank == 0) then
          write(*,'(a8,2a11)')"ustar","nitr","eitr"
          write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
       endif
#endif

       eitr = epsres; nitr = nmy
       call cgbase123pl &
            (amty,bvcy,xvcy,eitr,nitr,istat &
            ,amtypc,diagiy,ista2,iend1,jsta2v,jend1v,ksta2,kend1 &
            ,isolvers(3),iprecy,prec_name)
#ifdef VERBOSE
       if (myrank == 0) then
          write(*,'(a8,2a11)')"vstar","nitr","eitr"
          write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
       endif
#endif

       eitr = epsres; nitr = nmz
       call cgbase123pl &
            (amtz,bvcz,xvcz,eitr,nitr,istat &
            ,amtzpc,diagiz,ista2,iend1,jsta2,jend1,ksta2w,kend1w &
            ,isolvers(4),iprecz,prec_name)
#ifdef VERBOSE
       if (myrank == 0) then
          write(*,'(a8,2a11)')"wstar","nitr","eitr"
          write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
       endif
#endif

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista3,iend1
                uwk(i,j,k) = xvcx(i-1,j,k)*arelu + uwk(i,j,k)*arelu1
             enddo
          enddo
       enddo
       do k = ksta2,kend1
          do j = jsta3,jend1
             do i = ista2,iend1
                vwk(i,j,k) = xvcy(i,j-1,k)*arelv + vwk(i,j,k)*arelv1
             enddo
          enddo
       enddo
       do k = ksta3,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                wwk(i,j,k) = xvcz(i,j,k-1)*arelw + wwk(i,j,k)*arelw1
             enddo
          enddo
       enddo

       call velcndl(uwk,vwk,wwk,ivlt,ivlb,ista,iend,jsta,jend,ksta,kend)

       ! --------------------------------
       !     step 5: correct pressure
       ! --------------------------------

       call mypartialshift1l2(uwk,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift2l2(vwk,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift3l2(wwk,ista,iend,jsta,jend,ksta,kend)
       call calc_divvl &
            (uwk,vwk,wwk,divv,dx,dy,dz,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                bvcp(i,j,k) = divv(i,j,k)
                xvcp(i,j,k) = 0.e0_dp
             enddo
          enddo
       enddo
       if (myrank == 0)bvcp(2,2,2) = 0.e0_dp
       eitr = epsres
       nitr = nmp
       call cgbase123pl &
            (amtp,bvcp,xvcp,eitr,nitr,istat &
            ,amtppc,diagip,ista2,iend1,jsta2,jend1,ksta2,kend1 &
            ,isolvers(5),iprecp,prec_name)
#ifdef VERBOSE
       if (myrank == 0) then
          write(*,'(a8,2a11)')"pdash","nitr","eitr"
          write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
       endif
#endif

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                pdash(i,j,k) = xvcp(i,j,k)
             enddo
          enddo
       enddo

       ! --------------------------------------
       !     step 6: correct velocity field
       ! --------------------------------------

       call mypartialshift1l1(pdash,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift2l1(pdash,ista,iend,jsta,jend,ksta,kend)
       call mypartialshift3l1(pdash,ista,iend,jsta,jend,ksta,kend)

       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista3,iend1
                uwk(i,j,k) = uwk(i,j,k) + &
                     dudp1(i,j,k)*(pdash(i-1,j,k)-pdash(i,j,k))
             enddo
          enddo
       enddo

       do k = ksta2,kend1
          do j = jsta3,jend1
             do i = ista2,iend1
                vwk(i,j,k) = vwk(i,j,k) + &
                     dvdp1(i,j,k)*(pdash(i,j-1,k)-pdash(i,j,k))
             enddo
          enddo
       enddo

       do k = ksta3,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                wwk(i,j,k) = wwk(i,j,k) + &
                     dwdp1(i,j,k)*(pdash(i,j,k-1)-pdash(i,j,k))
             enddo
          enddo
       enddo

       ! ------------------------
       !    convergency check
       ! ------------------------
       call myroundshift123l(uwk,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)
       call myroundshift123l(vwk,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)
       call myroundshift123l(wwk,ista,iend,jsta,jend,ksta,kend &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1)
       call velcndl(uwk,vwk,wwk,ivlt,ivlb,ista,iend,jsta,jend,ksta,kend)
       vnorm2p = 0.e0_dp; divnorm2p = 0.e0_dp; vdhnorm2p = 0.e0_dp
       do k = ksta2,kend1
          do j = jsta2,jend1
             do i = ista2,iend1
                vnorm2p = vnorm2p+dx(i)*dy(j)*dz(k)* & ! vnorm2=<|v|^2>
                     ((uwk(i,j,k)+uwk(i+1,j,k))**2 &
                     +(vwk(i,j,k)+vwk(i,j+1,k))**2 &
                     +(wwk(i,j,k)+wwk(i,j,k+1))**2)
                divnorm2p = divnorm2p+dx(i)*dy(j)*dz(k)* & !divnorm2=<|div(v)|^2>
                     ((uwk(i+1,j,k)-uwk(i,j,k))*dxi(i) &
                     +(vwk(i,j+1,k)-vwk(i,j,k))*dyi(j) &
                     +(wwk(i,j,k+1)-wwk(i,j,k))*dzi(k))**2
                udash2 = uwk(i+1,j,k)+uwk(i,j,k)-uold(i+1,j,k)-uold(i,j,k)
                vdash2 = vwk(i,j+1,k)+vwk(i,j,k)-vold(i,j+1,k)-vold(i,j,k)
                wdash2 = wwk(i,j,k+1)+wwk(i,j,k)-wold(i,j,k+1)-wold(i,j,k)
                vdhnorm2p = vdhnorm2p+dx(i)*dy(j)*dz(k)* & ! vdhnorm2=<|v'|^2>
                     (udash2**2+vdash2**2+wdash2**2)
             end do
          end do
       end do
       rtmp1(1) = 0.25e0*vnorm2p
       rtmp1(2) = divnorm2p
       rtmp1(3) = 0.25e0*vdhnorm2p
       call mpi_allgather(rtmp1,3,mpi_double_precision, &
            rtmp2,3,mpi_double_precision,mpi_comm_world,ierr)
       vnorm2   = sum(rtmp2(1:3*nprocs:3))
       divnorm2 = sum(rtmp2(2:3*nprocs:3))
       vdhnorm2  =sum(rtmp2(3:3*nprocs:3))
       divzansa = dxyzmax*sqrt(divnorm2/vnorm2)
       vdhzansa = sqrt(vdhnorm2/vnorm2)
       zansa = max(divzansa,vdhzansa)
       if (ivis == 0) zansa = divzansa
       if (myrank == 0) then
          write(*,'(i5,4e12.4)')nsmp,zansa,&
               dxyzmax*sqrt(divnorm2),sqrt(vdhnorm2),sqrt(vnorm2)
       end if
       errors(nsmp) = zansa
       if (zansa <= epssim) then
          if (nsmp /= 1) then
             ich = 0
             exit
          end if
       end if

       ! relaxation is applied after 2nd iteration loop
       arelu = relu; arelu1 = 1.e0_dp-arelu
       arelv = relv; arelv1 = 1.e0_dp-arelv
       arelw = relw; arelw1 = 1.e0_dp-arelw
       arelp = relp; arelp1 = 1.e0_dp-arelp

    enddo simplerloop

    ! /////////////////////////
    !     simpler loop end
    ! /////////////////////////

    if (myrank == 0) then
       if (ich == 0) then
          write(*,50)nsmp,zansa
       else
          write(*,51)maxitr,zansa
       endif
    endif
50  format("OK/SIM ",i4,e11.4)
51  format("NG/SIM ",i4,e11.4)

    ! ---------------------------------
    !     replace array before exit
    ! ---------------------------------

    ul(ista:iend,jsta:jend,ksta:kend) = uwk(ista:iend,jsta:jend,ksta:kend)
    vl(ista:iend,jsta:jend,ksta:kend) = vwk(ista:iend,jsta:jend,ksta:kend)
    wl(ista:iend,jsta:jend,ksta:kend) = wwk(ista:iend,jsta:jend,ksta:kend)
    pl(ista:iend,jsta:jend,ksta:kend) = pwk(ista:iend,jsta:jend,ksta:kend)

    ! ------------------------------
    !     deallocate work arrays
    ! ------------------------------

    deallocate &
         (amtx,bvcx,xvcx,amtxpc,diagix &
         ,amty,bvcy,xvcy,amtypc,diagiy &
         ,amtz,bvcz,xvcz,amtzpc,diagiz &
         ,amtp,bvcp,xvcp,amtppc,diagip &
         ,stat=istat)
    !  write(*,'("Deallocation in hdesim3",i9)')istat
    return

  end subroutine hdesim3main

  subroutine calc_divvl &
       (ul,vl,wl,divvl,dx,dy,dz,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
    !*********************************************************************
    !     calculate div(v)*dV
    !*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    real(dp) :: divvl(ista:iend,jsta:jend,ksta:kend)
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: ul,vl,wl
    real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)

    do k = ksta2,kend1
       do j = jsta2,jend1
          do i = ista2,iend1
             divvl(i,j,k) = &
                  -dy(j)*dz(k)*(ul(i+1,j,k)-ul(i,j,k)) &
                  -dz(k)*dx(i)*(vl(i,j+1,k)-vl(i,j,k)) &
                  -dx(i)*dy(j)*(wl(i,j,k+1)-wl(i,j,k))
          enddo
       enddo
    enddo

    return
  end subroutine calc_divvl

  subroutine calvisg(vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy, &
       dx,dy,dz,dxs,dys,dzs,dxi,dyi,dzi,dxsi,dysi,dzsi &
       ,ista,iend,jsta,jend,ksta,kend)
    !*********************************************************************
    !     calculate viscosity at grid points
    !*********************************************************************
    integer i,j,k
    real(dp),intent(in) :: vcc(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    real(dp),intent(out) :: vgxy(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgxz(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgyx(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgyz(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgzx(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgzy(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    real(dp),intent(in):: dx(iax),dy(jax),dz(kax),dxs(iax),dys(jax),dzs(kax) &
         ,dxi(iax),dyi(jax),dzi(kax),dxsi(iax),dysi(jax),dzsi(kax)
    real(dp) :: dum1,dum2
    real(dp) :: vcci(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)

    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer :: ista2,iend1,jsta2,jend1,ksta2,kend1

    ista2 = max(2,ista); iend1 = min(iax-1,iend)
    jsta2 = max(2,jsta); jend1 = min(jax-1,jend)
    ksta2 = max(2,ksta); kend1 = min(kax-1,kend)

    do k = ksta2-1,kend1+1
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             vcci(i,j,k) = 1.e0_dp/vcc(i,j,k)
          enddo
       enddo
    enddo

    do k = ksta2,kend1
       do j = jsta2,jend1+1
          do i = ista2,iend
             dum1 = dy(j-1)*vcci(i-1,j-1,k)+dy(j)*vcci(i-1,j,k)
             dum2 = dy(j-1)*vcci(i,j-1,k)+dy(j)*vcci(i,j,k)
             vgxy(i,j,k) = dys(j)*dxsi(i)*(dx(i-1)/dum1+dx(i)/dum2)
          enddo
       enddo
    enddo

    do k = ksta2,kend1+1
       do j = jsta2,jend1
          do i = ista2,iend
             dum1 = dz(k-1)*vcci(i-1,j,k-1)+dz(k)*vcci(i-1,j,k)
             dum2 = dz(k-1)*vcci(i,j,k-1)+dz(k)*vcci(i,j,k)
             vgxz(i,j,k) = dzs(k)*dxsi(i)*(dx(i-1)/dum1+dx(i)/dum2)
          enddo
       enddo
    enddo

    do k = ksta2,kend1
       do j = jsta2,jend
          do i = ista2,iend1+1
             dum1 = dx(i-1)*vcci(i-1,j-1,k)+dx(i)*vcci(i,j-1,k)
             dum2 = dx(i-1)*vcci(i-1,j,k)+dx(i)*vcci(i,j,k)
             vgyx(i,j,k) = dxs(i)*dysi(j)*(dy(j-1)/dum1+dy(j)/dum2)
          enddo
       enddo
    enddo

    do k = ksta2,kend1+1
       do j = jsta2,jend
          do i = ista2,iend1
             dum1 = dz(k-1)*vcci(i,j-1,k-1)+dz(k)*vcci(i,j-1,k)
             dum2 = dz(k-1)*vcci(i,j,k-1)+dz(k)*vcci(i,j,k)
             vgyz(i,j,k) = dzs(k)*dysi(j)*(dy(j-1)/dum1+dy(j)/dum2)
          enddo
       enddo
    enddo

    do k = ksta2,kend
       do j = jsta2,jend1
          do i = ista2,iend1+1
             dum1 = dx(i-1)*vcci(i-1,j,k-1)+dx(i)*vcci(i,j,k-1)
             dum2 = dx(i-1)*vcci(i-1,j,k)+dx(i)*vcci(i,j,k)
             vgzx(i,j,k) = dxs(i)*dzsi(k)*(dz(k-1)/dum1+dz(k)/dum2)
          enddo
       enddo
    enddo

    do k = ksta2,kend
       do j = jsta2,jend1+1
          do i = ista2,iend1
             dum1 = dy(j-1)*vcci(i,j-1,k-1)+dy(j)*vcci(i,j,k-1)
             dum2 = dy(j-1)*vcci(i,j-1,k)+dy(j)*vcci(i,j,k)
             vgzy(i,j,k) = dys(j)*dzsi(k)*(dz(k-1)/dum1+dz(k)/dum2)
          enddo
       enddo
    enddo

    return
  end subroutine calvisg

  subroutine do_simpler_step2 &
       (uhatl,vhatl,whatl,uwkl,vwkl,wwkl,apx1l,apy1l,apz1l,dx,dy,dz &
       ,ivis,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl &
       ,ista,iend,jsta,jend,ksta,kend)
    !*********************************************************************
    !*********************************************************************

    integer,intent(in) :: ivis
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: uhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,whatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    real(dp),intent(in) :: uwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,wwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    real(dp),intent(in) :: apx1l(7,ista:iend,jsta:jend,ksta:kend) &
         ,apy1l(7,ista:iend,jsta:jend,ksta:kend) &
         ,apz1l(7,ista:iend,jsta:jend,ksta:kend)
    real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
    real(dp),intent(in) :: vgxyl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgxzl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgyxl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgyzl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgzxl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,vgzyl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,buoyl(ista:iend,jsta:jend,ksta:kend)
    integer ista2,ista3,iend1
    integer jsta2,jsta3,jend1
    integer ksta2,ksta3,kend1

    ista2 = max(2,ista); ista3 = max(3,ista); iend1 = min(iax-1,iend)
    jsta2 = max(2,jsta); jsta3 = max(3,jsta); jend1 = min(jax-1,jend)
    ksta2 = max(2,ksta); ksta3 = max(3,ksta); kend1 = min(kax-1,kend)

    if (ivis == 0) then
       call do_simpler_step2_constant_eta &
            (uhatl,vhatl,whatl,uwkl,vwkl,wwkl,apx1l,apy1l,apz1l,buoyl &
            ,ista,iend,jsta,jend,ksta,kend &
            ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
    else
       call do_simpler_step2_variable_eta &
            (uhatl,vhatl,whatl,uwkl,vwkl,wwkl,apx1l,apy1l,apz1l &
            ,dx,dy,dz,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl &
            ,ista,iend,jsta,jend,ksta,kend &
            ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
    endif

    return

  contains
    ! --------------------------------------------------------------------
    subroutine do_simpler_step2_constant_eta &
         (uhatl,vhatl,whatl,uwkl,vwkl,wwkl,apx1l,apy1l,apz1l,buoyl &
         ,ista,iend,jsta,jend,ksta,kend &
         ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
      !*********************************************************************
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp):: uhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,whatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
      real(dp),intent(in) :: uwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,wwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
      real(dp),intent(in) :: apx1l(7,ista:iend,jsta:jend,ksta:kend) &
           ,apy1l(7,ista:iend,jsta:jend,ksta:kend) &
           ,apz1l(7,ista:iend,jsta:jend,ksta:kend)
      real(dp),intent(in) :: buoyl(ista:iend,jsta:jend,ksta:kend)
      integer :: ista2,ista3,iend1
      integer :: jsta2,jsta3,jend1
      integer :: ksta2,ksta3,kend1

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista3,iend1
               uhatl(i,j,k) = &
                    (apx1l(1,i,j,k)*uwkl(i,j,k-1)+apx1l(2,i,j,k)*uwkl(i,j-1,k) &
                    +apx1l(3,i,j,k)*uwkl(i-1,j,k)+apx1l(5,i,j,k)*uwkl(i+1,j,k) &
                    +apx1l(6,i,j,k)*uwkl(i,j+1,k)+apx1l(7,i,j,k)*uwkl(i,j,k+1) &
                    )/apx1l(4,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta3,jend1
            do i = ista2,iend1
               vhatl(i,j,k) = &
                    (apy1l(1,i,j,k)*vwkl(i,j,k-1)+apy1l(2,i,j,k)*vwkl(i,j-1,k) &
                    +apy1l(3,i,j,k)*vwkl(i-1,j,k)+apy1l(5,i,j,k)*vwkl(i+1,j,k) &
                    +apy1l(6,i,j,k)*vwkl(i,j+1,k)+apy1l(7,i,j,k)*vwkl(i,j,k+1) &
                    )/apy1l(4,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta3,kend1
         do j = jsta2,jend1
            do i = ista2,iend1
               whatl(i,j,k) = &
                    (apz1l(1,i,j,k)*wwkl(i,j,k-1)+apz1l(2,i,j,k)*wwkl(i,j-1,k) &
                    +apz1l(3,i,j,k)*wwkl(i-1,j,k)+apz1l(5,i,j,k)*wwkl(i+1,j,k) &
                    +apz1l(6,i,j,k)*wwkl(i,j+1,k)+apz1l(7,i,j,k)*wwkl(i,j,k+1) &
                    +buoyl(i,j,k))/apz1l(4,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine do_simpler_step2_constant_eta
    !
    ! --------------------------------------------------------------------
    !
    subroutine do_simpler_step2_variable_eta &
         (uhatl,vhatl,whatl,uwkl,vwkl,wwkl,apx1l,apy1l,apz1l,dx,dy,dz &
         ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl &
         ,ista,iend,jsta,jend,ksta,kend &
         ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
      !*********************************************************************
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp) :: uhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vhatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,whatl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
      real(dp),intent(in) :: uwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,wwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
      real(dp),intent(in) :: apx1l(7,ista:iend,jsta:jend,ksta:kend) &
           ,apy1l(7,ista:iend,jsta:jend,ksta:kend) &
           ,apz1l(7,ista:iend,jsta:jend,ksta:kend)
      real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
      real(dp),intent(in) :: vgxyl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vgxzl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vgyxl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vgyzl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vgzxl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,vgzyl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           ,buoyl(ista:iend,jsta:jend,ksta:kend)
      integer :: ista2,ista3,iend1
      integer :: jsta2,jsta3,jend1
      integer :: ksta2,ksta3,kend1

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista3,iend1
               uhatl(i,j,k) =  &
                    (apx1l(1,i,j,k)*uwkl(i,j,k-1)+apx1l(2,i,j,k)*uwkl(i,j-1,k) &
                    +apx1l(3,i,j,k)*uwkl(i-1,j,k)+apx1l(5,i,j,k)*uwkl(i+1,j,k) &
                    +apx1l(6,i,j,k)*uwkl(i,j+1,k)+apx1l(7,i,j,k)*uwkl(i,j,k+1) &
                    +dz(k)*( &
                    vgxyl(i,j+1,k)*(vwkl(i,j+1,k)-vwkl(i-1,j+1,k)) &
                    -vgxyl(i,j,k)*(vwkl(i,j,k)-vwkl(i-1,j,k))) &
                    +dy(j)*( &
                    vgxzl(i,j,k+1)*(wwkl(i,j,k+1)-wwkl(i-1,j,k+1)) &
                    -vgxzl(i,j,k)*(wwkl(i,j,k)-wwkl(i-1,j,k))) &
                    )/apx1l(4,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta3,jend1
            do i = ista2,iend1
               vhatl(i,j,k) = &
                    (apy1l(1,i,j,k)*vwkl(i,j,k-1)+apy1l(2,i,j,k)*vwkl(i,j-1,k) &
                    +apy1l(3,i,j,k)*vwkl(i-1,j,k)+apy1l(5,i,j,k)*vwkl(i+1,j,k) &
                    +apy1l(6,i,j,k)*vwkl(i,j+1,k)+apy1l(7,i,j,k)*vwkl(i,j,k+1) &
                    +dz(k)*( &
                    vgyxl(i+1,j,k)*(uwkl(i+1,j,k)-uwkl(i+1,j-1,k)) &
                    -vgyxl(i,j,k)*(uwkl(i,j,k)-uwkl(i,j-1,k))) &
                    +dx(i)*( &
                    vgyzl(i,j,k+1)*(wwkl(i,j,k+1)-wwkl(i,j-1,k+1)) &
                    -vgyzl(i,j,k)*(wwkl(i,j,k)-wwkl(i,j-1,k))) &
                    )/apy1l(4,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta3,kend1
         do j = jsta2,jend1
            do i = ista2,iend1
               whatl(i,j,k) = &
                    (apz1l(1,i,j,k)*wwkl(i,j,k-1)+apz1l(2,i,j,k)*wwkl(i,j-1,k) &
                    +apz1l(3,i,j,k)*wwkl(i-1,j,k)+apz1l(5,i,j,k)*wwkl(i+1,j,k) &
                    +apz1l(6,i,j,k)*wwkl(i,j+1,k)+apz1l(7,i,j,k)*wwkl(i,j,k+1) &
                    +dy(j)*( &
                    vgzxl(i+1,j,k)*(uwkl(i+1,j,k)-uwkl(i+1,j,k-1)) &
                    -vgzxl(i,j,k)*(uwkl(i,j,k)-uwkl(i,j,k-1))) &
                    +dx(i)*( &
                    vgzyl(i,j+1,k)*(vwkl(i,j+1,k)-vwkl(i,j+1,k-1)) &
                    -vgzyl(i,j,k)*(vwkl(i,j,k)-vwkl(i,j,k-1))) &
                    +buoyl(i,j,k))/apz1l(4,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine do_simpler_step2_variable_eta
  end subroutine do_simpler_step2

  subroutine make_simpler_coefs_level1 &
       (ivis,apx1,apy1,apz1,app1,dudp1,dvdp1,dwdp1 &
       ,vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
       ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
       ,ista,iend,jsta,jend,ksta,kend)
    !*********************************************************************
    !     calculate coefficients for velocity at level 1
    !     only discretization is performed
    !*********************************************************************
    integer,intent(in) :: ivis
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer :: ista2,iend1,jsta2,jend1,ksta2,kend1
    real(dp),dimension(7,ista:iend,jsta:jend,ksta:kend) :: apx1,apy1,apz1,app1
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
         dudp1,dvdp1,dwdp1
    real(dp),intent(in),dimension(iax) :: dx,dxs,dxi,dxsi
    real(dp),intent(in),dimension(jax) :: dy,dys,dyi,dysi
    real(dp),intent(in),dimension(kax) :: dz,dzs,dzi,dzsi
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
         vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy
    !
    ista2 = max(2,ista); iend1 = min(iax-1,iend)
    jsta2 = max(2,jsta); jend1 = min(jax-1,jend)
    ksta2 = max(2,ksta); kend1 = min(kax-1,kend)
    !
    ! ----------------------------------------
    !     calculate coefficients for u/v/w
    ! ----------------------------------------
    !
    if (ivis == 0) then
       call velcoef_level1_constant_eta &
            (apx1,apy1,apz1 &
            ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
            ,ista,iend,jsta,jend,ksta,kend)
    else
       call velcoef_level1_variable_eta &
            (apx1,apy1,apz1,vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
            ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
            ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
            ,ista,iend,jsta,jend,ksta,kend)
    endif
    !
    ! ---------------------------------------------------
    !     calculate correction coefficients for u/v/w
    ! ---------------------------------------------------
    !
    call dvelcoef_level1 &
         (apx1,apy1,apz1,dx,dy,dz,dudp1,dvdp1,dwdp1 &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
    !
    ! -----------------------------------
    !     calculate coefficient for p
    ! -----------------------------------
    !
    call mypartialshift1l2(dudp1,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift2l2(dvdp1,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift3l2(dwdp1,ista,iend,jsta,jend,ksta,kend)
    call pcoef_level1 &
         (dudp1,dvdp1,dwdp1,dx,dy,dz,app1 &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
    !
    return

  contains
    ! --------------------------------------------------------------------
    !
    subroutine velcoef_level1_constant_eta &
         (apx1,apy1,apz1 &
         ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
      !*********************************************************************
      !     calculate coefficients for velocity at level 1
      !     for constant viscosity case
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp),dimension(7,ista:iend,jsta:jend,ksta:kend) :: apx1,apy1,apz1
      real(dp),intent(in),dimension(iax) :: dx,dxs,dxi,dxsi
      real(dp),intent(in),dimension(jax) :: dy,dys,dyi,dysi
      real(dp),intent(in),dimension(kax) :: dz,dzs,dzi,dzsi

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista2,iend
               apx1(1,i,j,k) = dxs(i)*dy(j)*dzsi(k)
               apx1(2,i,j,k) = dxs(i)*dysi(j)*dz(k)
               apx1(3,i,j,k) = dxi(i-1)*dy(j)*dz(k)
               apx1(5,i,j,k) = dxi(i)*dy(j)*dz(k)
               apx1(6,i,j,k) = dxs(i)*dysi(j+1)*dz(k)
               apx1(7,i,j,k) = dxs(i)*dy(j)*dzsi(k+1)
               apx1(4,i,j,k) = apx1(1,i,j,k)+apx1(2,i,j,k) &
                    +apx1(3,i,j,k)+apx1(5,i,j,k) &
                    +apx1(6,i,j,k)+apx1(7,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta2,jend
            do i = ista2,iend1
               apy1(1,i,j,k) = dx(i)*dys(j)*dzsi(k)
               apy1(2,i,j,k) = dx(i)*dyi(j-1)*dz(k)
               apy1(3,i,j,k) = dxsi(i)*dys(j)*dz(k)
               apy1(5,i,j,k) = dxsi(i+1)*dys(j)*dz(k)
               apy1(6,i,j,k) = dx(i)*dyi(j)*dz(k)
               apy1(7,i,j,k) = dx(i)*dys(j)*dzsi(k+1)
               apy1(4,i,j,k) = apy1(1,i,j,k)+apy1(2,i,j,k) &
                    +apy1(3,i,j,k)+apy1(5,i,j,k) &
                    +apy1(6,i,j,k)+apy1(7,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend
         do j = jsta2,jend1
            do i = ista2,iend1
               apz1(1,i,j,k) = dx(i)*dy(j)*dzi(k-1)
               apz1(2,i,j,k) = dx(i)*dysi(j)*dzs(k)
               apz1(3,i,j,k) = dxsi(i)*dy(j)*dzs(k)
               apz1(5,i,j,k) = dxsi(i+1)*dy(j)*dzs(k)
               apz1(6,i,j,k) = dx(i)*dysi(j+1)*dzs(k)
               apz1(7,i,j,k) = dx(i)*dy(j)*dzi(k)
               apz1(4,i,j,k) = apz1(1,i,j,k)+apz1(2,i,j,k) &
                    +apz1(3,i,j,k)+apz1(5,i,j,k) &
                    +apz1(6,i,j,k)+apz1(7,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine velcoef_level1_constant_eta
    !
    ! --------------------------------------------------------------------
    !
    subroutine velcoef_level1_variable_eta &
         (apx1,apy1,apz1,vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy &
         ,dx,dxs,dxi,dxsi,dy,dys,dyi,dysi,dz,dzs,dzi,dzsi &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
      !*********************************************************************
      !     calculate coefficients for velocity at level 1
      !     for variable viscosity case
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp),dimension(7,ista:iend,jsta:jend,ksta:kend) :: apx1,apy1,apz1
      real(dp),intent(in),dimension(iax) :: dx,dxs,dxi,dxsi
      real(dp),intent(in),dimension(jax) :: dy,dys,dyi,dysi
      real(dp),intent(in),dimension(kax) :: dz,dzs,dzi,dzsi
      real(dp),intent(in),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           :: vcc,vgxy,vgxz,vgyx,vgyz,vgzx,vgzy

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista2,iend
               apx1(1,i,j,k) = vgxz(i,j,k)*dxs(i)*dy(j)*dzsi(k)
               apx1(2,i,j,k) = vgxy(i,j,k)*dxs(i)*dysi(j)*dz(k)
               apx1(3,i,j,k) = 2.e0_dp*vcc(i-1,j,k)*dxi(i-1)*dy(j)*dz(k)
               apx1(5,i,j,k) = 2.e0_dp*vcc(i,j,k)*dxi(i)*dy(j)*dz(k)
               apx1(6,i,j,k) = vgxy(i,j+1,k)*dxs(i)*dysi(j+1)*dz(k)
               apx1(7,i,j,k) = vgxz(i,j,k+1)*dxs(i)*dy(j)*dzsi(k+1)
               apx1(4,i,j,k) = apx1(1,i,j,k)+apx1(2,i,j,k) &
                    +apx1(3,i,j,k)+apx1(5,i,j,k) &
                    +apx1(6,i,j,k)+apx1(7,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta2,jend
            do i = ista2,iend1
               apy1(1,i,j,k) = vgyz(i,j,k)*dx(i)*dys(j)*dzsi(k)
               apy1(2,i,j,k) = 2.e0_dp*vcc(i,j-1,k)*dx(i)*dyi(j-1)*dz(k)
               apy1(3,i,j,k) = vgyx(i,j,k)*dxsi(i)*dys(j)*dz(k)
               apy1(5,i,j,k) = vgyx(i+1,j,k)*dxsi(i+1)*dys(j)*dz(k)
               apy1(6,i,j,k) = 2.e0_dp*vcc(i,j,k)*dx(i)*dyi(j)*dz(k)
               apy1(7,i,j,k) = vgyz(i,j,k+1)*dx(i)*dys(j)*dzsi(k+1)
               apy1(4,i,j,k) = apy1(1,i,j,k)+apy1(2,i,j,k) &
                    +apy1(3,i,j,k)+apy1(5,i,j,k) &
                    +apy1(6,i,j,k)+apy1(7,i,j,k)
            enddo
         enddo
      enddo

      do k = ksta2,kend
         do j = jsta2,jend1
            do i = ista2,iend1
               apz1(1,i,j,k) = 2.e0_dp*vcc(i,j,k-1)*dx(i)*dy(j)*dzi(k-1)
               apz1(2,i,j,k) = vgzy(i,j,k)*dx(i)*dysi(j)*dzs(k)
               apz1(3,i,j,k) = vgzx(i,j,k)*dxsi(i)*dy(j)*dzs(k)
               apz1(5,i,j,k) = vgzx(i+1,j,k)*dxsi(i+1)*dy(j)*dzs(k)
               apz1(6,i,j,k) = vgzy(i,j+1,k)*dx(i)*dysi(j+1)*dzs(k)
               apz1(7,i,j,k) = 2.e0_dp*vcc(i,j,k)*dx(i)*dy(j)*dzi(k)
               apz1(4,i,j,k) = apz1(1,i,j,k)+apz1(2,i,j,k) &
                    +apz1(3,i,j,k)+apz1(5,i,j,k) &
                    +apz1(6,i,j,k)+apz1(7,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine velcoef_level1_variable_eta
    !
    ! --------------------------------------------------------------------
    !
    subroutine dvelcoef_level1 &
         (apx1,apy1,apz1,dx,dy,dz,dudp1,dvdp1,dwdp1 &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
      !*********************************************************************
      !     calculate velocity correction due to pressure change
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp),intent(in),dimension(7,ista:iend,jsta:jend,ksta:kend) :: &
           apx1,apy1,apz1
      real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
      real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
           dudp1,dvdp1,dwdp1

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista2,iend
               dudp1(i,j,k) = dy(j)*dz(k)/apx1(4,i,j,k)
            enddo
         enddo
      enddo
      do k = ksta2,kend1
         do j = jsta2,jend
            do i = ista2,iend1
               dvdp1(i,j,k) = dz(k)*dx(i)/apy1(4,i,j,k)
            enddo
         enddo
      enddo
      do k = ksta2,kend
         do j = jsta2,jend1
            do i = ista2,iend1
               dwdp1(i,j,k) = dx(i)*dy(j)/apz1(4,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine dvelcoef_level1
    !
    ! --------------------------------------------------------------------
    !
    subroutine pcoef_level1 &
         (dudp1,dvdp1,dwdp1,dx,dy,dz,app1 &
         ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
         ,ista,iend,jsta,jend,ksta,kend)
      !*********************************************************************
      !     calculate coefficient for pressure
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      real(dp),intent(in),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
           :: dudp1,dvdp1,dwdp1
      real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
      real(dp),dimension(7,ista:iend,jsta:jend,ksta:kend) :: app1
      !
      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista2,iend1
               app1(1,i,j,k) = dx(i)*dy(j)*dwdp1(i,j,k)
               app1(2,i,j,k) = dz(k)*dx(i)*dvdp1(i,j,k)
               app1(3,i,j,k) = dy(j)*dz(k)*dudp1(i,j,k)
               app1(5,i,j,k) = dy(j)*dz(k)*dudp1(i+1,j,k)
               app1(6,i,j,k) = dz(k)*dx(i)*dvdp1(i,j+1,k)
               app1(7,i,j,k) = dx(i)*dy(j)*dwdp1(i,j,k+1)
               app1(4,i,j,k) = app1(1,i,j,k)+app1(2,i,j,k) &
                    +app1(3,i,j,k)+app1(5,i,j,k) &
                    +app1(6,i,j,k)+app1(7,i,j,k)
            enddo
         enddo
      enddo

      return
    end subroutine pcoef_level1

  end subroutine make_simpler_coefs_level1

  subroutine make_simpler_coefs_level2 &
       (apx1,apy1,apz1,app1,apx2,apy2,apz2,app2,dudp1,dvdp1,dwdp1 &
       ,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1,ivlt,ivlb)
    !*********************************************************************
    !     impose boundary conditions for coefficients of u/v/w/p
    !*********************************************************************
    integer :: i,j,k,ij
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    real(dp) :: apx1(7,ista:iend,jsta:jend,ksta:kend) &
         ,apx2(7,ista:iend,jsta:jend,ksta:kend)
    real(dp) :: apy1(7,ista:iend,jsta:jend,ksta:kend) &
         ,apy2(7,ista:iend,jsta:jend,ksta:kend)
    real(dp) :: apz1(7,ista:iend,jsta:jend,ksta:kend) &
         ,apz2(7,ista:iend,jsta:jend,ksta:kend)
    real(dp) :: app1(7,ista:iend,jsta:jend,ksta:kend) &
         ,app2(7,ista:iend,jsta:jend,ksta:kend)
    real(dp) :: dudp1(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,dvdp1(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         ,dwdp1(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    integer ivlt,ivlb
    real(dp) at,ab
    !
    !     set boundary conditions for top/bottom surface
    !
    if (ivlb == 1) then
       ab = 1.e0_dp
    else if (ivlb == 2) then
       ab = -1.e0_dp
    endif
    if (ivlt == 1) then
       at = 1.e0_dp
    else if (ivlt == 2) then
       at = -1.e0_dp
    endif

    apx2 = apx1
    apy2 = apy1
    apz2 = apz1
    app2 = app1

    ! bottom boundary
    k = 2
    ! impermeable, free slip or rigid surface
    !
    if (myrankk == 0) then

       do j = jsta2,jend1
          do i = ista2,iend
             apx2(4,i,j,k) = apx2(4,i,j,k) - ab*apx2(1,i,j,k)
             apx2(1,i,j,k) = 0.e0_dp
          enddo
       enddo

       do j = jsta2,jend
          do i = ista2,iend1
             apy2(4,i,j,k) = apy2(4,i,j,k) - ab*apy2(1,i,j,k)
             apy2(1,i,j,k) = 0.e0_dp
          enddo
       enddo

       do j = jsta2,jend1
          do i = ista2,iend1
             apz2(1,i,j,k+1) = 0.e0_dp
          enddo
       enddo

       do j = jsta2,jend1
          do i = ista2,iend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(1,i,j,k)
             app2(1,i,j,k) = 0.e0_dp
          enddo
       enddo

       do j = jsta2,jend1
          do i = ista2,iend1
             dwdp1(i,j,k) = 0.e0_dp
          enddo
       enddo

    endif
    !
    ! top boundary
    k = kax-1
    ! impermeable, free slip or rigid surface
    !
    if (myrankk == kprocs-1) then
       !
       do j = jsta2,jend1
          do i = ista2,iend
             apx2(4,i,j,k) = apx2(4,i,j,k) - at*apx2(7,i,j,k)
             apx2(7,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do j = jsta2,jend
          do i = ista2,iend1
             apy2(4,i,j,k) = apy2(4,i,j,k) - at*apy2(7,i,j,k)
             apy2(7,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do j = jsta2,jend1
          do i = ista2,iend1
             apz2(7,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do j = jsta2,jend1
          do i = ista2,iend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(7,i,j,k)
             app2(7,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do j = jsta2,jend1
          do i = ista2,iend1
             dwdp1(i,j,k+1) = 0.e0_dp
          enddo
       enddo
       !
    endif
    !
    ! side wall
    i = 2
    ! impermeable free slip
    !
    if (myranki == 0) then
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             apx2(3,i+1,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend
             apy2(4,i,j,k) = apy2(4,i,j,k) - apy2(3,i,j,k)
             apy2(3,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend
          do j = jsta2,jend1
             apz2(4,i,j,k) = apz2(4,i,j,k) - apz2(3,i,j,k)
             apz2(3,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(3,i,j,k)
             app2(3,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             dudp1(i,j,k) = 0.e0_dp
          enddo
       enddo
       !
    endif
    !
    ! side wall
    i = iax-1
    ! impermeable free slip
    !
    if (myranki == iprocs-1) then
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             apx2(5,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend
             apy2(4,i,j,k) = apy2(4,i,j,k) - apy2(5,i,j,k)
             apy2(5,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend
          do j = jsta2,jend1
             apz2(4,i,j,k) = apz2(4,i,j,k) - apz2(5,i,j,k)
             apz2(5,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(5,i,j,k)
             app2(5,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do j = jsta2,jend1
             dudp1(i+1,j,k) = 0.e0_dp
          enddo
       enddo
       !
    endif
    !
    ! side wall
    j = 2
    ! impermeable free slip
    !
    if (myrankj == 0) then
       !
       do k = ksta2,kend1
          do i = ista2,iend
             apx2(4,i,j,k) = apx2(4,i,j,k) - apx2(2,i,j,k)
             apx2(2,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do i = ista2,iend1
             apy2(2,i,j+1,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend
          do i = ista2,iend1
             apz2(4,i,j,k) = apz2(4,i,j,k) - apz2(2,i,j,k)
             apz2(2,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do i = ista2,iend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(2,i,j,k)
             app2(2,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do i = ista2,iend1
             dvdp1(i,j,k) = 0.e0_dp
          enddo
       enddo
       !
    endif
    !
    ! side wall
    j = jax-1
    ! impermeable free slip
    !
    if (myrankj == jprocs-1) then
       !
       do k = ksta2,kend1
          do i = ista2,iend
             apx2(4,i,j,k) = apx2(4,i,j,k) - apx2(6,i,j,k)
             apx2(6,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend1
          do i = ista2,iend1
             apy2(6,i,j,k) = 0.e0_dp
          enddo
       enddo
       !
       do k = ksta2,kend
          do i = ista2,iend1
             apz2(4,i,j,k) = apz2(4,i,j,k) - apz2(6,i,j,k)
             apz2(6,i,j,k) = 0.e0_dp
          enddo
       enddo
!
       do k = ksta2,kend1
          do i = ista2,iend1
             app2(4,i,j,k) = app2(4,i,j,k) - app2(6,i,j,k)
             app2(6,i,j,k) = 0.e0_dp
          enddo
       enddo
!
       do k = ksta2,kend1
          do i = ista2,iend1
             dvdp1(i,j+1,k) = 0.e0_dp
          enddo
       enddo
       !
    endif
    !
    !     to avoid singularity set p(2,2,2)=0
    !
    if (myrank == 0) then
       app2(1,2,2,2) = 0.e0_dp
       app2(2,2,2,2) = 0.e0_dp
       app2(3,2,2,2) = 0.e0_dp
       app2(5,2,2,2) = 0.e0_dp
       app2(6,2,2,2) = 0.e0_dp
       app2(7,2,2,2) = 0.e0_dp
       app2(3,3,2,2) = 0.e0_dp
       app2(2,2,3,2) = 0.e0_dp
       app2(1,2,2,3) = 0.e0_dp
    endif
    !
    return
  end subroutine make_simpler_coefs_level2

  subroutine make_simpler_matrice &
       (apx2l,apy2l,apz2l,app2l,amtxl,amtyl,amtzl,amtpl &
       ,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
       ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w)
    !*********************************************************************
    !*********************************************************************
    integer :: i,j,k,ij
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    integer,intent(in) :: ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w
    integer :: ista3,jsta3,ksta3
    real(dp),intent(in),dimension(7,ista:iend,jsta:jend,ksta:kend) :: &
         apx2l,apy2l,apz2l,app2l
    real(dp) :: amtxl(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
         ,amtyl(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1,7) &
         ,amtzl(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1,7) &
         ,amtpl(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7)
    !
    ista3 = max(3,ista); jsta3 = max(3,jsta); ksta3 = max(3,ksta)
    !
    !     ---- x ----
    !
    amtxl = 0.e0_dp
    do k = ksta2,kend1
       do j = jsta2,jend1
          do i = ista3,iend1
             amtxl(i-1,j,k,1) = -apx2l(1,i,j,k)
             amtxl(i-1,j,k,2) = -apx2l(2,i,j,k)
             amtxl(i-1,j,k,3) = -apx2l(3,i,j,k)
             amtxl(i-1,j,k,4) = apx2l(4,i,j,k)
             amtxl(i-1,j,k,5) = -apx2l(5,i,j,k)
             amtxl(i-1,j,k,6) = -apx2l(6,i,j,k)
             amtxl(i-1,j,k,7) = -apx2l(7,i,j,k)
          enddo
       enddo
    enddo
    !
    !     ---- y ----
    !
    amtyl = 0.e0_dp
    do k = ksta2,kend1
       do j = jsta3,jend1
          do i = ista2,iend1
             amtyl(i,j-1,k,1) = -apy2l(1,i,j,k)
             amtyl(i,j-1,k,2) = -apy2l(2,i,j,k)
             amtyl(i,j-1,k,3) = -apy2l(3,i,j,k)
             amtyl(i,j-1,k,4) = apy2l(4,i,j,k)
             amtyl(i,j-1,k,5) = -apy2l(5,i,j,k)
             amtyl(i,j-1,k,6) = -apy2l(6,i,j,k)
             amtyl(i,j-1,k,7) = -apy2l(7,i,j,k)
          enddo
       enddo
    enddo
    !
    !     ---- z ----
    !
    amtzl = 0.e0_dp
    do k = ksta3,kend1
       do j = jsta2,jend1
          do i = ista2,iend1
             amtzl(i,j,k-1,1) = -apz2l(1,i,j,k)
             amtzl(i,j,k-1,2) = -apz2l(2,i,j,k)
             amtzl(i,j,k-1,3) = -apz2l(3,i,j,k)
             amtzl(i,j,k-1,4) = apz2l(4,i,j,k)
             amtzl(i,j,k-1,5) = -apz2l(5,i,j,k)
             amtzl(i,j,k-1,6) = -apz2l(6,i,j,k)
             amtzl(i,j,k-1,7) = -apz2l(7,i,j,k)
          enddo
       enddo
    enddo
    !
    !     ---- p ----
    !
    amtpl = 0.e0_dp
    do k = ksta2,kend1
       do j = jsta2,jend1
          do i = ista2,iend1
             amtpl(i,j,k,1) = -app2l(1,i,j,k)
             amtpl(i,j,k,2) = -app2l(2,i,j,k)
             amtpl(i,j,k,3) = -app2l(3,i,j,k)
             amtpl(i,j,k,4) = app2l(4,i,j,k)
             amtpl(i,j,k,5) = -app2l(5,i,j,k)
             amtpl(i,j,k,6) = -app2l(6,i,j,k)
             amtpl(i,j,k,7) = -app2l(7,i,j,k)
          enddo
       enddo
    enddo
    !
    return
  end subroutine make_simpler_matrice

  subroutine prep_simpler_step4 &
       (bvcxl,bvcyl,bvczl,uwkl,vwkl,wwkl,dx,dy,dz,ivis &
       ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl,pwkl &
       ,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
       ,istau,iendu,jstav,jendv,kstaw,kendw &
       ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w)
    !*********************************************************************
    !*********************************************************************
    integer,intent(in) :: ivis
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    integer,intent(in) :: istau,iendu,jstav,jendv,kstaw,kendw
    integer,intent(in) :: ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w
    real(dp) :: bvcxl(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
         ,bvcyl(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
         ,bvczl(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1)
    real(dp),intent(in),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) &
         :: uwkl,vwkl,wwkl,pwkl &
         ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl

    real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
    real(dp),intent(in) :: buoyl(ista:iend,jsta:jend,ksta:kend)
    integer :: ista3,jsta3,ksta3

    ista3 = max(3,ista); jsta3 = max(3,jsta); ksta3 = max(3,ksta)

    if (ivis == 0) then
       call prep_simpler_step4_constant_eta &
            (bvcxl,bvcyl,bvczl,dx,dy,dz,buoyl,pwkl &
            ,ista,iend,jsta,jend,ksta,kend &
            ,istau,iendu,jstav,jendv,kstaw,kendw &
            ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w &
            ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
    else
       call prep_simpler_step4_variable_eta &
            (bvcxl,bvcyl,bvczl,uwkl,vwkl,wwkl,dx,dy,dz &
            ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl,pwkl &
            ,ista,iend,jsta,jend,ksta,kend &
            ,istau,iendu,jstav,jendv,kstaw,kendw &
            ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w &
            ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
    endif
    !
    return

  contains
    ! --------------------------------------------------------------------

    subroutine prep_simpler_step4_constant_eta &
         (bvcxl,bvcyl,bvczl,dx,dy,dz,buoyl,pwkl &
         ,ista,iend,jsta,jend,ksta,kend &
         ,istau,iendu,jstav,jendv,kstaw,kendw &
         ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w &
         ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
      !*********************************************************************
      !*********************************************************************
      integer :: i,j,k
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
      integer,intent(in) :: istau,iendu,jstav,jendv,kstaw,kendw
      integer,intent(in) :: ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w
      real(dp) bvcxl(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
           ,bvcyl(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
           ,bvczl(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1)
      real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
      real(dp),intent(in) :: buoyl(ista:iend,jsta:jend,ksta:kend) &
           ,pwkl(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
      integer,intent(in) :: ista3,jsta3,ksta3

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista3,iend1
               bvcxl(i-1,j,k) = &
                    dy(j)*dz(k)*(pwkl(i-1,j,k)-pwkl(i,j,k))
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta3,jend1
            do i = ista2,iend1
               bvcyl(i,j-1,k) = &
                    dz(k)*dx(i)*(pwkl(i,j-1,k)-pwkl(i,j,k))
            enddo
         enddo
      enddo
      do k = ksta3,kend1
         do j = jsta2,jend1
            do i = ista2,iend1
               bvczl(i,j,k-1) = buoyl(i,j,k) &
                    +dx(i)*dy(j)*(pwkl(i,j,k-1)-pwkl(i,j,k))
            enddo
         enddo
      enddo

      return
    end subroutine prep_simpler_step4_constant_eta

    ! --------------------------------------------------------------------

    subroutine prep_simpler_step4_variable_eta &
         (bvcxl,bvcyl,bvczl,uwkl,vwkl,wwkl,dx,dy,dz &
         ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl,buoyl,pwkl &
         ,ista,iend,jsta,jend,ksta,kend &
         ,istau,iendu,jstav,jendv,kstaw,kendw &
         ,ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w &
         ,ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1)
      !*********************************************************************
      !*********************************************************************
      integer i,j,k
      integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
      integer,intent(in) :: istau,iendu,jstav,jendv,kstaw,kendw
      integer,intent(in) :: ista2u,iend1u,jsta2v,jend1v,ksta2w,kend1w
      real(dp) :: bvcxl(ista2u-1:iend1u+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
           ,bvcyl(ista2-1:iend1+1,jsta2v-1:jend1v+1,ksta2-1:kend1+1) &
           ,bvczl(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2w-1:kend1w+1)
      real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
           uwkl,vwkl,wwkl,pwkl &
           ,vgxyl,vgxzl,vgyxl,vgyzl,vgzxl,vgzyl
      real(dp),intent(in) :: dx(iax),dy(jax),dz(kax)
      real(dp),intent(in) :: buoyl(ista:iend,jsta:jend,ksta:kend)
      integer,intent(in) :: ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1

      do k = ksta2,kend1
         do j = jsta2,jend1
            do i = ista3,iend1
               bvcxl(i-1,j,k) = dz(k)*( &
                    vgxyl(i,j+1,k)*(vwkl(i,j+1,k)-vwkl(i-1,j+1,k)) &
                    -vgxyl(i,j,k)*(vwkl(i,j,k)-vwkl(i-1,j,k))) &
                    +dy(j)*( &
                    vgxzl(i,j,k+1)*(wwkl(i,j,k+1)-wwkl(i-1,j,k+1)) &
                    -vgxzl(i,j,k)*(wwkl(i,j,k)-wwkl(i-1,j,k))) &
                    +dy(j)*dz(k)*(pwkl(i-1,j,k)-pwkl(i,j,k))
            enddo
         enddo
      enddo

      do k = ksta2,kend1
         do j = jsta3,jend1
            do i = ista2,iend1
               bvcyl(i,j-1,k) = dz(k)*( &
                    vgyxl(i+1,j,k)*(uwkl(i+1,j,k)-uwkl(i+1,j-1,k)) &
                    -vgyxl(i,j,k)*(uwkl(i,j,k)-uwkl(i,j-1,k))) &
                    +dx(i)*( &
                    vgyzl(i,j,k+1)*(wwkl(i,j,k+1)-wwkl(i,j-1,k+1)) &
                    -vgyzl(i,j,k)*(wwkl(i,j,k)-wwkl(i,j-1,k))) &
                    +dz(k)*dx(i)*(pwkl(i,j-1,k)-pwkl(i,j,k))
            enddo
         enddo
      enddo
      !
      do k = ksta3,kend1
         do j = jsta2,jend1
            do i = ista2,iend1
               bvczl(i,j,k-1) = dy(j)*( &
                    vgzxl(i+1,j,k)*(uwkl(i+1,j,k)-uwkl(i+1,j,k-1)) &
                    -vgzxl(i,j,k)*(uwkl(i,j,k)-uwkl(i,j,k-1))) &
                    +dx(i)*( &
                    vgzyl(i,j+1,k)*(vwkl(i,j+1,k)-vwkl(i,j+1,k-1)) &
                    -vgzyl(i,j,k)*(vwkl(i,j,k)-vwkl(i,j,k-1))) &
                    +buoyl(i,j,k) &
                    +dx(i)*dy(j)*(pwkl(i,j,k-1)-pwkl(i,j,k))
            enddo
         enddo
      enddo

      return
    end subroutine prep_simpler_step4_variable_eta

  end subroutine prep_simpler_step4

  subroutine velcndl(ul,vl,wl,ivlt,ivlb &
       ,ista,iend,jsta,jend,ksta,kend)
    !***********************************************************************
    !     apply boundary conditions for velocity components
    !
    !     ivlt: =1 for shear-free top surface, =2 for rigid top surface
    !     ivlb: =1 for shear-free top bottom, =2 for rigid bottom surface
    !***********************************************************************
    integer :: i,j,k
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: ul,vl,wl
    integer,intent(in) :: ivlt,ivlb
    real(dp) :: at,ab
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer :: ista2,ista3,iend1,jsta2,jsta3,jend1,ksta2,ksta3,kend1

    !     boundary conditions

    if (ivlb == 1) then
       ab = 1.e0_dp
    else if (ivlb == 2) then
       ab = -1.e0_dp
    endif
    if (ivlt == 1) then
       at = 1.e0_dp
    else if (ivlt == 2) then
       at = -1.e0_dp
    endif
    !
    ista2 = max(2,ista); ista3 = max(3,ista); iend1 = min(iax-1,iend)
    jsta2 = max(2,jsta); jsta3 = max(3,jsta); jend1 = min(jax-1,jend)
    ksta2 = max(2,ksta); ksta3 = max(3,ksta); kend1 = min(kax-1,kend)
    !
    ! --------------------------------------
    !     apply boundary condition for u
    ! --------------------------------------
    !
    if (myranki == 0) then
       do k = ksta2,kend1
          do j = jsta2,jend1
             ul(2,j,k) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myranki == iprocs-1) then
       do k = ksta2,kend1
          do j = jsta2,jend1
             ul(iax,j,k) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myrankj == 0) then
       do k = ksta2,kend1
          do i = ista2,iend1
             ul(i,1,k) = ul(i,2,k)
          enddo
       enddo
    endif
    if (myrankj == jprocs-1) then
       do k = ksta2,kend1
          do i = ista2,iend1
             ul(i,jax,k) = ul(i,jax-1,k)
          enddo
       enddo
    endif
    !
    if (myrankk == 0) then
       do j = jsta,jend
          do i = ista2,iend
             ul(i,j,1) = ab*ul(i,j,2)
          enddo
       enddo
    endif
    if (myrankk == kprocs-1) then
       do j = jsta,jend
          do i = ista2,iend
             ul(i,j,kax) = at*ul(i,j,kax-1)
          enddo
       enddo
    endif
    !
    ! --------------------------------------
    !     apply boundary condition for v
    ! --------------------------------------
    !
    if (myrankj == 0) then
       do k = ksta2,kend1
          do i = ista2,iend1
             vl(i,2,k) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myrankj == jprocs-1) then
       do k = ksta2,kend1
          do i = ista2,iend1
             vl(i,jax,k) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myranki == 0) then
       do k = ksta2,kend1
          do j = jsta2,jend
             vl(1,j,k) = vl(2,j,k)
          enddo
       enddo
    endif
    !
    if (myranki == iprocs-1) then
       do k = ksta2,kend1
          do j = jsta2,jend
             vl(iax,j,k) = vl(iax-1,j,k)
          enddo
       enddo
    endif
    !
    if (myrankk == 0) then
       do j = jsta2,jend
          do i = ista,iend
             vl(i,j,1) = ab*vl(i,j,2)
          enddo
       enddo
    endif
    if (myrankk == kprocs-1) then
       do j = jsta2,jend
          do i = ista,iend
             vl(i,j,kax) = at*vl(i,j,kax-1)
          enddo
       enddo
    endif
    !
    ! --------------------------------------
    !     apply boundary condition for w
    ! --------------------------------------
    !
    if (myrankk == 0) then
       do j = jsta2,jend1
          do i = ista2,iend1
             wl(i,j,2) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myrankk == kprocs-1) then
       do j = jsta2,jend1
          do i = ista2,iend1
             wl(i,j,kax) = 0.e0_dp
          enddo
       enddo
    endif
    !
    if (myranki == 0) then
       do k = ksta2,kend1
          do j = jsta2,jend1
             wl(1,j,k) = wl(2,j,k)
          enddo
       enddo
    endif
    !
    if (myranki == iprocs-1) then
       do k = ksta2,kend1
          do j = jsta2,jend1
             wl(iax,j,k) = wl(iax-1,j,k)
          enddo
       enddo
    endif
    !
    if (myrankj == 0) then
       do k = ksta2,kend
          do i = ista,iend
             wl(i,1,k) = wl(i,2,k)
          enddo
       enddo
    endif
    !
    if (myrankj == jprocs-1) then
       do k = ksta2,kend
          do i = ista,iend
             wl(i,jax,k) = wl(i,jax-1,k)
          enddo
       enddo
    endif
    !
    return
  end subroutine velcndl

  subroutine viscnd(visc,vcc,ista,iend,jsta,jend,ksta,kend)
    !*********************************************************************
    !     set boundary conditions for viscosity
    !*********************************************************************
    integer :: i,j,k
    real(dp),intent(in) :: visc(ista:iend,jsta:jend,ksta:kend)
    real(dp) :: vcc(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    integer :: ista2,iend1,jsta2,jend1,ksta2,kend1

    ista2 = max(2,ista); iend1 = min(iax-1,iend)
    jsta2 = max(2,jsta); jend1 = min(jax-1,jend)
    ksta2 = max(2,ksta); kend1 = min(kax-1,kend)

    do k = ksta2,kend1
       do j = jsta2,jend1
          do i = ista2,iend1
             vcc(i,j,k) = visc(i,j,k)
          enddo
       enddo
    enddo

    if (myrankj == 0) then
       do k = ksta2,kend1
          do i = ista2,iend1
             vcc(i,1,k) = vcc(i,2,k)
          enddo
       enddo
    endif

    if (myrankj == jprocs-1) then
       do k = ksta2,kend1
          do i = ista2,iend1
             vcc(i,jax,k) = vcc(i,jax-1,k)
          enddo
       enddo
    endif

    if (myranki == 0) then
       do k = ksta2,kend1
          do j = jsta,jend
             vcc(1,j,k) = vcc(2,j,k)
          enddo
       enddo
    endif

    if (myranki == iprocs-1) then
       do k = ksta2,kend1
          do j = jsta,jend
             vcc(iax,j,k) = vcc(iax-1,j,k)
          enddo
       enddo
    endif

    if (myrankk == 0) then
       do j = jsta,jend
          do i = ista,iend
             vcc(i,j,1) = vcc(i,j,2)
          enddo
       enddo
    endif

    if (myrankk == kprocs-1) then
       do j = jsta,jend
          do i = ista,iend
             vcc(i,j,kax) = vcc(i,j,kax-1)
          enddo
       enddo
    endif

    return
  end subroutine viscnd

end module hdesim3
