subroutine htrns3e(tmpl,ul,vl,wl,schl,tsurf,tbot,qbot,vboti,dt)
  use constants; use mesh; use mympitools; use division
!**********************************************************************
!     solve energy equation by a power-law scheme
!     adopted boundary conditions are;
!        fixed temperature on top/bottom boundaries
!        zero heat flux across side walls
!
!     input
!        tmp: temperature at time step n
!        u,v,w: velocity
!        sch: heat generation
!        tsurf: surface temperature
!        tbot: bottom temperature at time step n
!        vboti: inverse of volume of heat bath at bottom
!        dx,dy,dz: mesh size
!        dt: time step
!     output
!        tmp: temperature at time step n+1
!        tbot: bottom temperature at time step n+1
!        qbot: bottom heat flux
!     local
!        pe: Peclet number
!        twk: work array for temperature
!        fx,fy,fz: heat flux in x/y/z direction
!*********************************************************************
  implicit none

  integer :: i,j,k
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,schl,ul,vl,wl
  real(dp),intent(in) :: dt
  real(dp) :: tsurf,tbot,qbot,vboti,tb,tb1
  real(dp) :: pe,faca,facb,dum,dum1
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: twk,fx,fy,fz
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: difx,dify,difz

! ---------------------------------------
!     temperature change in heat bath
! ---------------------------------------

  if (myrankk == 0) then
     tb = 0.e0_dp
     do j = jsta2,jend1
        do i = ista2,iend1
           tb = tb + tmpl(i,j,2)*dx(i)*dy(j)
        enddo
     enddo
  else
     tb = 0.e0_dp
  endif
  call mpi_allreduce(tb,tb1,1,mpi_double_precision,mpi_sum &
       ,mpi_comm_world,ierr)
  tb = tb1/aspx/aspy
  qbot = 2.e0_dp*(tbot-tb)/dz(2)
  tbot = tbot - vboti*qbot*dt

! -----------------------
!     set work arrays
! -----------------------

  twk = 0.e0_dp
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,k) = tmpl(i,j,k)
        enddo
     enddo
  enddo

!     top/bottom boundary
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,1) = 2.e0_dp*tbot-tmpl(i,j,2)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,kax) = 2.e0_dp*tsurf-tmpl(i,j,kax-1)
        enddo
     enddo
  endif

!     side walls
  if (myranki == 0) then
     do k = ksta,kend
        do j = jsta2,jend1
           twk(1,j,k) = twk(2,j,k)
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta,kend
        do j = jsta2,jend1
           twk(iax,j,k) = twk(iax-1,j,k)
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta,kend
        do i = ista,iend
           twk(i,1,k) = twk(i,2,k)
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta,kend
        do i = ista,iend
           twk(i,jax,k) = twk(i,jax-1,k)
        enddo
     enddo
  endif

! -------------------
!     conductance
! -------------------
  difx = 0.e0_dp; dify = 0.e0_dp; difz = 0.e0_dp
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend
           difx(i,j,k) = 1.e0_dp/dxs(i)
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend
        do i = ista2,iend1
           dify(i,j,k) = 1.e0_dp/dys(j)
        enddo
     enddo
  enddo
  do k = ksta2,kend
     do j = jsta2,jend1
        do i = ista2,iend1
           difz(i,j,k) = 1.e0_dp/dzs(k)
        enddo
     enddo
  enddo

! ------------------------------
!     define total heat flux
!     using power-law scheme
! ------------------------------
  call mypartialshift1l1(twk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l1(twk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l1(twk,ista,iend,jsta,jend,ksta,kend)

  fx = 0.e0_dp; fy = 0.e0_dp; fz = 0.e0_dp

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend
           pe = ul(i,j,k)/difx(i,j,k); dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(0.e0_dp,dum**5)+dmax1(0.e0_dp,-pe); facb = faca + pe
           fx(i,j,k) = difx(i,j,k)*(facb*twk(i-1,j,k)-faca*twk(i,j,k))
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend
        do i = ista2,iend1
           pe = vl(i,j,k)/dify(i,j,k); dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(0.e0_dp,dum**5)+dmax1(0.e0_dp,-pe); facb = faca + pe
           fy(i,j,k) = dify(i,j,k)*(facb*twk(i,j-1,k)-faca*twk(i,j,k))
        enddo
     enddo
  enddo
  do k = ksta2,kend
     do j = jsta2,jend1
        do i = ista2,iend1
           pe = wl(i,j,k)/difz(i,j,k); dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(0.e0_dp,dum**5)+dmax1(0.e0_dp,-pe); facb = faca + pe
           fz(i,j,k) = difz(i,j,k)*(facb*twk(i,j,k-1)-faca*twk(i,j,k))
        enddo
     enddo
  enddo

! --------------------------
!     evolve temperature
! --------------------------

  call mypartialshift1l2(fx,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(fy,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(fz,ista,iend,jsta,jend,ksta,kend)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,k) = twk(i,j,k) &
                +dt*(schl(i,j,k) &
                +(fx(i,j,k)-fx(i+1,j,k))/dx(i) &
                +(fy(i,j,k)-fy(i,j+1,k))/dy(j) &
                +(fz(i,j,k)-fz(i,j,k+1))/dz(k))
        enddo
     enddo
  enddo

! ---------------------
!     replace array
! ---------------------

!     top/bottom boundary
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,1) = 2.e0_dp*tbot-tmpl(i,j,2)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,kax) = 2.e0_dp*tsurf-tmpl(i,j,kax-1)
        enddo
     enddo
  endif

!     side walls
  if (myranki == 0) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(1,j,k) = tmpl(2,j,k)
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(iax,j,k) = tmpl(iax-1,j,k)
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,1,k) = tmpl(i,2,k)
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,jax,k) = tmpl(i,jax-1,k)
        enddo
     enddo
  endif

  return
end subroutine htrns3e
