subroutine htrns3e1(tmpl,ul,vl,wl,schl,tsurf,tbot,qbot,vboti,dt)
  use constants; use mesh; use mympitools; use division
!**********************************************************************
!     solve energy equation
!     1st-order Euler time integration scheme is used
!     total heat flux is calculated by power-law scheme
!     adopted boundary conditions are;
!        fixed temperature on top/bottom boundaries
!        zero heat flux across side walls
!
!     input
!        tmp: temperature at time step n
!        u,v,w: velocity
!        sch: heat generation
!        tsurf: surface temperature
!        tbot: bottom temperature at time step n
!        vboti: inverse of volume of heat bath at bottom
!        dx,dy,dz: mesh size
!        dt: time step
!     output
!        tmp: temperature at time step n+1
!        tbot: bottom temperature at time step n+1
!        qbot: bottom heat flux
!     local
!        pe: Peclet number
!        twk: work array for temperature
!        fx,fy,fz: heat flux in x/y/z direction
!*********************************************************************
  implicit none

  integer :: i,j,k
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,schl,ul,vl,wl
  real(dp),intent(in) :: dt
  real(dp) :: tsurf,tbot,qbot,vboti,tb,tb1
  real(dp) :: pe,faca,facb,difx,dify,difz,dum
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: twk,fx,fy,fz

  if (vboti > zero) then
! ---------------------------------------
!     temperature change in heat bath
! ---------------------------------------
     if (myrankk == 0) then
        tb = zero
        do j = jsta2,jend1
           do i = ista2,iend1
              tb = tb + tmpl(i,j,2)*dx(i)*dy(j)
           enddo
        enddo
     else
        tb = zero
     endif
     call mpi_allreduce(tb,tb1,1,mpi_double_precision,mpi_sum &
          ,mpi_comm_world,ierr)
     tb = tb1/aspx/aspy
     qbot = 2.e0_dp*(tbot-tb)*dzi(2)
     tbot = tbot - vboti*qbot*dt
  end if

! -----------------------
!     set work arrays
! -----------------------

  twk = zero
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,k) = tmpl(i,j,k)
        enddo
     enddo
  enddo

!     top/bottom boundary
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,1) = 2.e0_dp*tbot-tmpl(i,j,2)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           twk(i,j,kax) = 2.e0_dp*tsurf-tmpl(i,j,kax-1)
        enddo
     enddo
  endif

!     side walls
  if (myranki == 0) then
     do k = ksta,kend
        do j = jsta2,jend1
           twk(1,j,k) = twk(2,j,k)
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta,kend
        do j = jsta2,jend1
           twk(iax,j,k) = twk(iax-1,j,k)
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta,kend
        do i = ista,iend
           twk(i,1,k) = twk(i,2,k)
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta,kend
        do i = ista,iend
           twk(i,jax,k) = twk(i,jax-1,k)
        enddo
     enddo
  endif

! ------------------------------
!     define total heat flux
!     using power-law scheme
! ------------------------------
  call mypartialshift1l1(twk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l1(twk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l1(twk,ista,iend,jsta,jend,ksta,kend)

  fx = zero; fy = zero; fz = zero

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend
           difx = 1.e0_dp*dxsi(i)
           pe = ul(i,j,k)/difx; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           fx(i,j,k) = difx*(facb*twk(i-1,j,k)-faca*twk(i,j,k))
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend
        dify = 1.e0_dp*dysi(j)
        do i = ista2,iend1
           pe = vl(i,j,k)/dify; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           fy(i,j,k) = dify*(facb*twk(i,j-1,k)-faca*twk(i,j,k))
        enddo
     enddo
  enddo
  do k = ksta2,kend
     difz = 1.e0_dp*dzsi(k)
     do j = jsta2,jend1
        do i = ista2,iend1
           pe = wl(i,j,k)/difz; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           fz(i,j,k) = difz*(facb*twk(i,j,k-1)-faca*twk(i,j,k))
        enddo
     enddo
  enddo

! --------------------------
!     evolve temperature
! --------------------------

  call mypartialshift1l2(fx,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(fy,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(fz,ista,iend,jsta,jend,ksta,kend)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,k) = twk(i,j,k) &
                +dt*(schl(i,j,k) &
                +(fx(i,j,k)-fx(i+1,j,k))*dxi(i) &
                +(fy(i,j,k)-fy(i,j+1,k))*dyi(j) &
                +(fz(i,j,k)-fz(i,j,k+1))*dzi(k))
        enddo
     enddo
  enddo

! ---------------------
!     replace array
! ---------------------

!     top/bottom boundary
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,1) = 2.e0_dp*tbot-tmpl(i,j,2)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,kax) = 2.e0_dp*tsurf-tmpl(i,j,kax-1)
        enddo
     enddo
  endif

!     side walls
  if (myranki == 0) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(1,j,k) = tmpl(2,j,k)
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(iax,j,k) = tmpl(iax-1,j,k)
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,1,k) = tmpl(i,2,k)
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,jax,k) = tmpl(i,jax-1,k)
        enddo
     enddo
  endif

  return
end subroutine htrns3e1
