subroutine htrns3e2(tmpl,ul,vl,wl,schl,tsurf,tbot,qbot,vboti,dt)
  use constants; use mesh; use mympitools; use division
!**********************************************************************
!     solve energy equation
!     2nd-order Euler time integration scheme is used
!     total heat flux is calculated by power-law scheme
!     adopted boundary conditions are;
!        fixed temperature on top/bottom boundaries
!        zero heat flux across side walls
!
!     input
!        tmp: temperature at time step n
!        u,v,w: velocity
!        sch: heat generation
!        tsurf: surface temperature
!        tbot: bottom temperature at time step n
!        vboti: inverse of volume of heat bath at bottom
!        dx,dy,dz: mesh size
!        dt: time step
!     output
!        tmp: temperature at time step n+1
!        tbot: bottom temperature at time step n+1
!        qbot: bottom heat flux
!     local
!        pe: Peclet number
!        twk0/twk1/twk2: work array for temperature
!        tincr0/tincr1: work array for temperature increment
!        difx,dify,difz: conductance in x/y/z direction
!*********************************************************************
  implicit none

  integer :: i,j,k,ij
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,schl,ul,vl,wl
  real(dp),intent(in) :: dt,tsurf,vboti
  real(dp) :: dt2,tbot,qbot,tb,tb1,pe,faca,facb,difx,dify,difz,dum
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
       cax,cay,caz,cbx,cby,cbz,twk0,twk1,twk2
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend,7) :: amt
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: bvt,tincr0,tincr1
  real(dp) :: amt4dtmax

  if (vboti > zero) then
  ! ---------------------------------------
  !     temperature change in heat bath
  ! ---------------------------------------
     if (myrankk == 0) then
        tb = zero
        do j = jsta2,jend1
           do i = ista2,iend1
              tb = tb + tmpl(i,j,2)*dx(i)*dy(j)
           enddo
        enddo
     else
        tb = zero
     endif
     call mpi_allreduce(tb,tb1,1,mpi_double_precision,mpi_sum &
          ,mpi_comm_world,ierr)
     tb = tb1/aspx/aspy
     qbot = 2.e0_dp*(tbot-tb)/dz(2)
     tbot = tbot - vboti*qbot*dt
  end if

  ! ----------------------------------------
  !     calculate heat flux coefficients
  ! ----------------------------------------
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend
           difx = 1.e0_dp/dxs(i)
           pe = ul(i,j,k)/difx; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           cax(i,j,k) = difx*faca
           cbx(i,j,k) = difx*facb
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend
        dify = 1.e0_dp/dys(j)
        do i = ista2,iend1
           pe = vl(i,j,k)/dify; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           cay(i,j,k) = dify*faca
           cby(i,j,k) = dify*facb
        enddo
     enddo
  enddo
  do k = ksta2,kend
     difz = 1.e0_dp/dzs(k)
     do j = jsta2,jend1
        do i = ista2,iend1
           pe = wl(i,j,k)/difz; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           caz(i,j,k) = difz*faca
           cbz(i,j,k) = difz*facb
        enddo
     enddo
  enddo

  ! ----------------------------------------------
  !     calculate coefficients for temperature
  ! ----------------------------------------------
  call mypartialshift1l2(cax,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift1l2(cbx,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(cay,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(cby,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(caz,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(cbz,ista,iend,jsta,jend,ksta,kend)

  amt = zero; bvt = zero; dt2 = 0.5e0_dp*dt
  twk0 = zero; twk1 = zero; twk2 = zero; tincr0 = zero; tincr1 = zero

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           amt(i,j,k,1) = -cbz(i,j,k)/dz(k)
           amt(i,j,k,2) = -cby(i,j,k)/dy(j)
           amt(i,j,k,3) = -cbx(i,j,k)/dx(i)
           amt(i,j,k,5) = -cax(i+1,j,k)/dx(i)
           amt(i,j,k,6) = -cay(i,j+1,k)/dy(j)
           amt(i,j,k,7) = -caz(i,j,k+1)/dz(k)
           amt(i,j,k,4) = &
                +(cax(i,j,k)+cbx(i+1,j,k))/dx(i) &
                +(cay(i,j,k)+cby(i,j+1,k))/dy(j) &
                +(caz(i,j,k)+cbz(i,j,k+1))/dz(k)
           bvt(i,j,k) = schl(i,j,k)
           twk0(i,j,k) = tmpl(i,j,k)
        enddo
     enddo
  enddo
  amt4dtmax = maxval(dt*amt(ista2:iend1,jsta2:jend1,ksta2:kend1,4))
  if (amt4dtmax >= 1.e0_dp) write(*,*)"1st-order Euler would fail.",amt4dtmax

  ! calculate temperature increment 0
  call myshift123l(twk0,ista,iend,jsta,jend,ksta,kend)
  call tmp_bound(twk0,tsurf,tbot)
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tincr0(i,j,k) = bvt(i,j,k) &
                -amt(i,j,k,1)*twk0(i,j,k-1) &
                -amt(i,j,k,2)*twk0(i,j-1,k) &
                -amt(i,j,k,3)*twk0(i-1,j,k) &
                -amt(i,j,k,4)*twk0(i,j,k) &
                -amt(i,j,k,5)*twk0(i+1,j,k) &
                -amt(i,j,k,6)*twk0(i,j+1,k) &
                -amt(i,j,k,7)*twk0(i,j,k+1)
        end do
     end do
  end do

  ! calculate intermediate temperature 1
  ! equivalent to solution with 1st-order Euler scheme
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           twk1(i,j,k) = tmpl(i,j,k)+dt*tincr0(i,j,k)
        end do
     end do
  end do

  ! calculate temperature increment 1
  call myshift123l(twk1,ista,iend,jsta,jend,ksta,kend)
  call tmp_bound(twk1,tsurf,tbot)
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tincr1(i,j,k) = bvt(i,j,k) &
                -amt(i,j,k,1)*twk1(i,j,k-1) &
                -amt(i,j,k,2)*twk1(i,j-1,k) &
                -amt(i,j,k,3)*twk1(i-1,j,k) &
                -amt(i,j,k,4)*twk1(i,j,k) &
                -amt(i,j,k,5)*twk1(i+1,j,k) &
                -amt(i,j,k,6)*twk1(i,j+1,k) &
                -amt(i,j,k,7)*twk1(i,j,k+1)
        end do
     end do
  end do

  ! update temperature
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           twk2(i,j,k) = twk0(i,j,k)+dt2*(tincr0(i,j,k)+tincr1(i,j,k))
        end do
     end do
  end do
  call tmp_bound(twk2,tsurf,tbot)

  ! replace array
  do k = ksta,kend
     do j = jsta,jend
        do i = ista,iend
           tmpl(i,j,k) = twk2(i,j,k) ! 2nd-order Euler
        end do
     end do
  end do
  return

contains

  subroutine tmp_bound(twk,tsurf,tbot)
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: twk
    real(dp),intent(in) :: tsurf,tbot
    ! top/bottom boundary
    if (myrankk == 0) then
       do j = jsta2,jend1
          do i = ista2,iend1
             twk(i,j,1) = 2.e0_dp*tbot-twk(i,j,2)
          enddo
       enddo
    endif
    if (myrankk == kprocs-1) then
       do j = jsta2,jend1
          do i = ista2,iend1
             twk(i,j,kax) = 2.e0_dp*tsurf-twk(i,j,kax-1)
          enddo
       enddo
    endif
    !  side walls
    if (myranki == 0) then
       do k = ksta,kend
          do j = jsta2,jend1
             twk(1,j,k) = twk(2,j,k)
          enddo
       enddo
    endif
    if (myranki == iprocs-1) then
       do k = ksta,kend
          do j = jsta2,jend1
             twk(iax,j,k) = twk(iax-1,j,k)
          enddo
       enddo
    endif
    if (myrankj == 0) then
       do k = ksta,kend
          do i = ista,iend
             twk(i,1,k) = twk(i,2,k)
          enddo
       enddo
    endif
    if (myrankj == jprocs-1) then
       do k = ksta,kend
          do i = ista,iend
             twk(i,jax,k) = twk(i,jax-1,k)
          enddo
       enddo
    endif
    return
  end subroutine tmp_bound
end subroutine htrns3e2
