subroutine htrns3i(tmpl,ul,vl,wl,schl,tsurf,tbot,qbot,vboti,dt,isolver)
  use constants; use mesh; use mympitools; use division; use mylinsol
!**********************************************************************
!     solve energy equation by implicit, power-law scheme
!     adopted boundary conditions are;
!        fixed temperature on top/bottom boundaries
!        zero heat flux across side walls
!
!     simultaneous equation for convection-diffusion is solved by
!     bi-conjugate gradient method
!
!     ver.0.0 on 20030108
!     ver.0.1 on 20030109
!     ver.0.2 on 20030120
!     ver.1.0 on 20030207
!
!     input
!        tmp: temperature at time step n
!        u,v,w: velocity
!        sch: heat generation
!        tsurf: surface temperature
!        tbot: bottom temperature at time step n
!        vboti: inverse of volume of heat bath at bottom
!        dx,dy,dz: mesh size
!        dt: time step
!     output
!        tmp: temperature at time step n+1
!        tbot: bottom temperature at time step n+1
!        qbot: bottom heat flux
!     local
!        pe: Peclet number
!        twk: work array for temperature
!        fx,fy,fz: heat flux in x/z direction
!*********************************************************************
  implicit none

  integer :: i,j,k,ij
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,schl,ul,vl,wl
  real(dp),intent(in) :: dt,tsurf,vboti
  real(dp) :: dti,tbot,qbot,tb,tb1,pe,faca,facb,difx,dify,difz,dum
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
       cax,cay,caz,cbx,cby,cbz
  real(dp),allocatable :: at1(:,:,:,:),at1prec(:,:,:,:),bt1(:,:,:), &
       twk(:,:,:),diagit(:,:,:)

  ! variables for linear solver
  integer :: nitr,istat,iprec,isolver
  integer,parameter :: maxsor=9999
  real(dp) :: eitr
  real(dp),parameter :: eps=1.d-10
  character(4) :: prec_name

  if (vboti > zero) then
  ! ---------------------------------------
  !     temperature change in heat bath
  ! ---------------------------------------
     if (myrankk == 0) then
        tb = zero
        do j = jsta2,jend1
           do i = ista2,iend1
              tb = tb + tmpl(i,j,2)*dx(i)*dy(j)
           enddo
        enddo
     else
        tb = zero
     endif
     call mpi_allreduce(tb,tb1,1,mpi_double_precision,mpi_sum &
          ,mpi_comm_world,ierr)
     tb = tb1/aspx/aspy
     qbot = 2.e0_dp*(tbot-tb)/dz(2)
     tbot = tbot - vboti*qbot*dt
  end if

  ! ----------------------------------------
  !     calculate heat flux coefficients
  ! ----------------------------------------
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend
           difx = 1.e0_dp/dxs(i)
           pe = ul(i,j,k)/difx; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           cax(i,j,k) = difx*faca
           cbx(i,j,k) = difx*facb
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend
        dify = 1.e0_dp/dys(j)
        do i = ista2,iend1
           pe = vl(i,j,k)/dify; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           cay(i,j,k) = dify*faca
           cby(i,j,k) = dify*facb
        enddo
     enddo
  enddo
  do k = ksta2,kend
     difz = 1.e0_dp/dzs(k)
     do j = jsta2,jend1
        do i = ista2,iend1
           pe = wl(i,j,k)/difz; dum = 1.e0_dp-0.1e0_dp*dabs(pe)
           faca = dmax1(zero,dum**5)+dmax1(zero,-pe); facb = faca + pe
           caz(i,j,k) = difz*faca
           cbz(i,j,k) = difz*facb
        enddo
     enddo
  enddo

  ! ----------------------------------------------
  !     calculate coefficients for temperature
  ! ----------------------------------------------
  call mypartialshift1l2(cax,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift1l2(cbx,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(cay,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(cby,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(caz,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(cbz,ista,iend,jsta,jend,ksta,kend)
  allocate &
       (at1(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
       ,at1prec(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
       ,bt1(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
       ,twk(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
       ,diagit(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1))
  at1 = zero; bt1 = zero; twk = zero; dti = 1.e0_dp/dt
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           at1(i,j,k,1) = -cbz(i,j,k)/dz(k)
           at1(i,j,k,2) = -cby(i,j,k)/dy(j)
           at1(i,j,k,3) = -cbx(i,j,k)/dx(i)
           at1(i,j,k,5) = -cax(i+1,j,k)/dx(i)
           at1(i,j,k,6) = -cay(i,j+1,k)/dy(j)
           at1(i,j,k,7) = -caz(i,j,k+1)/dz(k)
           at1(i,j,k,4) = dti &
                +(cax(i,j,k)+cbx(i+1,j,k))/dx(i) &
                +(cay(i,j,k)+cby(i,j+1,k))/dy(j) &
                +(caz(i,j,k)+cbz(i,j,k+1))/dz(k)
           bt1(i,j,k) = tmpl(i,j,k)*dti + schl(i,j,k)
           twk(i,j,k) = tmpl(i,j,k)
        enddo
     enddo
  enddo

  ! ---------------------------
  !     boundary conditions
  ! ---------------------------
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           at1(i,j,2,4) = at1(i,j,2,4)-at1(i,j,2,1)
           bt1(i,j,2) = bt1(i,j,2)- 2.e0_dp*at1(i,j,2,1)*tbot
           at1(i,j,2,1) = zero
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           at1(i,j,kax-1,4) = at1(i,j,kax-1,4)-at1(i,j,kax-1,7)
           bt1(i,j,kax-1) = bt1(i,j,kax-1) &
                - 2.e0_dp*at1(i,j,kax-1,7)*tsurf
           at1(i,j,kax-1,7) = zero
        enddo
     enddo
  endif
  if (myranki == 0) then
     do k = ksta2,kend1
        do j = jsta2,jend1
           at1(2,j,k,4) = at1(2,j,k,4)+at1(2,j,k,3)
           at1(2,j,k,3) = zero
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta2,kend1
        do j = jsta2,jend1
           at1(iax-1,j,k,4) = at1(iax-1,j,k,4)+at1(iax-1,j,k,5)
           at1(iax-1,j,k,5) = zero
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta2,kend1
        do i = ista2,iend1
           at1(i,2,k,4) = at1(i,2,k,4)+at1(i,2,k,2)
           at1(i,2,k,2) = zero
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta2,kend1
        do i = ista2,iend1
           at1(i,jax-1,k,4) = at1(i,jax-1,k,4)+at1(i,jax-1,k,6)
           at1(i,jax-1,k,6) = zero
        enddo
     enddo
  endif

  ! ///////////////////////////
  !     begin linear solver
  ! ///////////////////////////
  nitr = maxsor; eitr = eps; iprec = 0
  !  call ilubcg123pl(at1,bt1,twk,nx,ny,nz,eitr,nitr,istat,iprec &
  !       ,at1prec,diagit,ista,iend,jsta,jend,ksta,kend)
  call bcgbase123pl(at1,bt1,twk,eitr,nitr,istat &
       ,at1prec,diagit,ista2,iend1,jsta2,jend1,ksta2,kend1 &
       ,isolver,iprec,prec_name)

  if (myrank == 0) then
     write(*,'(a8,2a11)')"temp.","nitr","eitr"
     write(*,'(a5,i3,i11,e11.4)')prec_name,istat,nitr,eitr
  endif
  ! /////////////////////////
  !     end linear solver
  ! /////////////////////////

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,k) = twk(i,j,k)
        enddo
     enddo
  enddo
  deallocate (at1,at1prec,bt1,twk,diagit)

  ! top/bottom boundary
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,1) = 2.e0_dp*tbot-tmpl(i,j,2)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tmpl(i,j,kax) = 2.e0_dp*tsurf-tmpl(i,j,kax-1)
        enddo
     enddo
  endif

  ! side walls
  if (myranki == 0) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(1,j,k) = tmpl(2,j,k)
        enddo
     enddo
  endif
  if (myranki == iprocs-1) then
     do k = ksta,kend
        do j = jsta2,jend1
           tmpl(iax,j,k) = tmpl(iax-1,j,k)
        enddo
     enddo
  endif
  if (myrankj == 0) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,1,k) = tmpl(i,2,k)
        enddo
     enddo
  endif
  if (myrankj == jprocs-1) then
     do k = ksta,kend
        do i = ista,iend
           tmpl(i,jax,k) = tmpl(i,jax-1,k)
        enddo
     enddo
  endif

  return
end subroutine htrns3i
