module mesh
  use constants
  implicit none
  include "meshnum.h"

  real(dp) :: aspx,aspy
  real(dp),dimension(iax) :: dx,dxs,xp,xc,dxi,dxsi
  real(dp),dimension(jax) :: dy,dys,yp,yc,dyi,dysi
  real(dp),dimension(kax) :: dz,dzs,zp,zc,dzi,dzsi
  real(dp) :: dxmax,dxmin,dymax,dymin,dzmax,dzmin,dxyzmax,dxyzmin

contains
  subroutine dfmesh
    !**********************************************************************
    !     define mesh system
    !
    !          dxs(I)
    !                dx(i)
    !        |   |     |    |     |
    !  J+1  -w---+-----w----+-----+-
    !        |   |     |    |     |
    !   j   -T---u-----T----u-----T- dz(j)
    !        |   |     |    |     |
    !   J   -w---+-----w----+-----+- dzs(J)
    !        |   |     |    |     |
    !  j-1  -T---u-----T----u-----T-
    !        |   |     |    |     |
    !
    !            I     i   I+1   i+1
    !
    !     dxs(I) = (dx(i-1)+dx(i))/2
    !     dzs(J) = (dz(j-1)+dz(j))/2
    !**********************************************************************
    implicit none

    integer :: i,j,k
!    real(dp),parameter :: rx=1.06e0_dp, ry=1.06e0_dp, rz=1.07e0_dp
    real(dp),parameter :: rx=1.e0_dp, ry=1.e0_dp, rz=1.e0_dp
    real(dp) :: dumx,dumy,dumz,sum,rr

! =================
!     define dx
! =================

    dx(2) = 1.e0_dp
    do i = 3,iax/2
       dx(i) = dx(i-1)*rx
    enddo
    dx(iax-1) = 1.e0_dp
    do i = iax-2,iax/2+1,-1
       dx(i) = dx(i+1)*rx
    enddo
    sum = 0.e0_dp
    do i = 2,iax-1
       sum = sum + dx(i)
    enddo
    rr = aspx/sum
    do i = 2,iax-1
       dx(i) = dx(i)*rr
    enddo

    dx(1)   = dx(2)
    dx(iax) = dx(iax-1)

! =================
!     define dy
! =================

    dy(2) = 1.e0_dp
    do j = 3,jax/2
       dy(j) = dy(j-1)*ry
    enddo
    dy(jax-1) = 1.e0_dp
    do j = jax-2,jax/2+1,-1
       dy(j) = dy(j+1)*ry
    enddo
    sum = 0.e0_dp
    do j = 2,jax-1
       sum = sum + dy(j)
    enddo
    rr = aspy/sum
    do j = 2,jax-1
       dy(j) = dy(j)*rr
    enddo

    dy(1)   = dy(2)
    dy(jax) = dy(jax-1)

! =================
!     define dz
! =================

    dz(2) = 1.e0_dp
    do k = 3,kax/2
       dz(k) = dz(k-1)*rz
    enddo
    dz(kax-1) = 1.e0_dp
    do k = kax-2,kax/2+1,-1
       dz(k) = dz(k+1)*rz
    enddo
    sum = 0.e0_dp
    do k = 2,kax-1
       sum = sum + dz(k)
    enddo
    rr = 1.e0_dp/sum
    do k = 2,kax-1
       dz(k) = dz(k)*rr
    enddo

    dz(1)   = dz(2)
    dz(kax) = dz(kax-1)

! ==================
!     define dxs
! ==================

    do i = 2,iax
       dxs(i) = 0.5e0_dp*(dx(i-1)+dx(i))
    enddo
    dxs(1) = dxs(2)

! ==================
!     define dys
! ==================

    do j = 2,jax
       dys(j) = 0.5e0_dp*(dy(j-1)+dy(j))
    enddo
    dys(1) = dys(2)

! ==================
!     define dzs
! ==================

    do k = 2,kax
       dzs(k) = 0.5e0_dp*(dz(k-1)+dz(k))
    enddo
    dzs(1) = dzs(2)

! ----------------------------
!     set coodinate system
! ----------------------------

    xp(2) = 0.5e0_dp*dx(2)
    xp(1) = xp(2)-dxs(2)
    do i = 3,iax
       xp(i) = xp(i-1)+dxs(i)
    enddo
    yp(2) = 0.5e0_dp*dy(2)
    yp(1) = yp(2)-dys(2)
    do j = 3,jax
       yp(j) = yp(j-1)+dys(j)
    enddo
    zp(2) = 0.5e0_dp*dz(2)
    zp(1) = zp(2)-dzs(2)
    do k = 3,kax
       zp(k) = zp(k-1)+dzs(k)
    enddo

    xc(1) = -dx(1)
    xc(2) = 0.e0_dp
    do i = 3,iax
       xc(i) = xc(i-1)+dx(i-1)
    enddo
    yc(1) = -dy(1)
    yc(2) = 0.e0_dp
    do j = 3,jax
       yc(j) = yc(j-1)+dy(j-1)
    enddo
    zc(1) = -dz(1)
    zc(2) = 0.e0_dp
    do k = 3,kax
       zc(k) = zc(k-1)+dz(k-1)
    enddo

    return
  end subroutine dfmesh

  subroutine mconst
    !**********************************************************************
    !     calculate mesh constants
    !**********************************************************************
    implicit none

    ! inverse of mesh size
    dxi = 1.e0_dp/dx; dxsi = 1.e0_dp/dxs
    dyi = 1.e0_dp/dy; dysi = 1.e0_dp/dys
    dzi = 1.e0_dp/dz; dzsi = 1.e0_dp/dzs

    ! find maximum/minimum mesh size
    dxmin = minval(dx); dxmax = maxval(dx)
    dymin = minval(dy); dymax = maxval(dy)
    dzmin = minval(dz); dzmax = maxval(dz)
    dxyzmin = min(dxmin,dymin,dzmin); dxyzmax = max(dxmax,dymax,dzmax)

    return
  end subroutine mconst
end module mesh
